/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include "MERLBRDFRead.h"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <assert.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_sf_erf.h>
#include <gsl/gsl_sf_gamma.h>
// This depends on the actual location of your levmar library: 
#include "../../../../../levmar-2.6/levmar.h"

// To compile:
// g++ -O3 -I /usr/local/include fit_HeModel.cpp -lm -lstdc++ -L /usr/local/lib -lgsl ../../../../../levmar-2.6/liblevmar.a -llapack -lcblas -o HeModel
// (depending on the actual location of GSL and levmar, adapt the above line)
//
// To run: ./HeModel /path/to/MERL/reflectance.binary > reflectance.out
//
#include "common_functions.inl"

// To compute z0, we need to solve the equation (z0/sigma0) = a * exp(-0.5 * (z_0/sigma_0)^2)

inline double f(double a, double u) { return u - a * exp(-0.5 * u * u); }
inline double fp(double a, double u) { return 1. + u * a * exp(-0.5 * u * u); }

double solveForZ0(double a) {
	// Find the roots using Newton's method
	double u0 = 0.5; 
	double fval = f(a, u0); 
	double fpval = fp(a, u0); 
	double oldu0 = HUGE_VAL;

	while ((fpval > 1e-5) && (fabs(oldu0 - u0) > 1e-5)) {
		oldu0 = u0; 
		u0 -= fval / fpval; 
		if (u0 < 0) u0 = 0; 
		fval = f(a, u0); 
		fpval = fp(a, u0); 
	}
	return u0; 
}



// Computes BRDF values but works on 3 channels at once. Saves computations.
void brdf_values(double result[3], double n[3],  double k[3], double diffuse[3], double sigma_0, double tau, double theta_h, double theta_d, double theta_i, double theta_o, double phi_i, double phi_o, double phi_d)
{
	//RP: Where is equation 70 implemented ???
	result[0] = 0.0;
	result[1] = 0.0;
	result[2] = 0.0;
  if (theta_o >= 0.5 * M_PI) return;
  if (theta_i >= 0.5 * M_PI) return;

  // Computing G (Geometric term) -- Equation 76
  // v . v / v_z 
  double v2_vz = 2 * cos(theta_d) / cos(theta_h);
  // Instead of computing cross product for sin, compute dot product and sin^2 = 1 - cos^2
  double c_io = sin(theta_o) * cos(phi_o) * sin(theta_i) * cos(phi_i) + 
  	  sin(theta_o) * sin(phi_o) * sin(theta_i) * sin(phi_i) + cos(theta_o) * cos(theta_i); 
  double s2_io = 1 - c_io * c_io;
  if (s2_io < 1e-7) s2_io = 1e-7; // ??? 
  // 
  double s_r_k_i = sin(theta_i) * sin(phi_i - phi_o + M_PI);    
  double s_i_k_r = sin(theta_o) * sin(phi_o - phi_i + M_PI);     
  double p_r_k_i = -(sin(theta_i) * cos(theta_o) * cos(phi_o - phi_i + M_PI) + cos(theta_i) * sin(theta_o));
  double p_i_k_r = sin(theta_o) * cos(theta_i) * cos(phi_o - phi_i + M_PI) + cos(theta_o) * sin(theta_i); 

  double G = (v2_vz * v2_vz /(s2_io * s2_io)) * (s_r_k_i * s_r_k_i + p_r_k_i * p_r_k_i) * (s_i_k_r * s_i_k_r + p_i_k_r * p_i_k_r);
  if (isnan(G) || isinf(G)) G = 0; 

  // Computing S (Shadowing term) -- Equation 77
  // Note: we reuse the erfc computations because this function is expensive
  // RP:  using std::erfc is preferable...or am I missing something ???
  double tau_i = tau/(2. * sigma_0 * tan(theta_i)); 
  double erfc_theta_i = gsl_sf_erfc(tau_i);
  if (isnan(erfc_theta_i) || isinf(erfc_theta_i)) erfc_theta_i = 0.;
  double tau_o = tau/(2. * sigma_0 * tan(theta_o)); 
  double erfc_theta_o = gsl_sf_erfc(tau_o);
  if (isnan(erfc_theta_o) || isinf(erfc_theta_o)) erfc_theta_o = 0.;
  double Lambda_i = 0.5 * (exp(- tau_i * tau_i)/(sqrt(M_PI) * tau_i) - erfc_theta_i); 
  double Lambda_o = 0.5 * (exp(- tau_o * tau_o)/(sqrt(M_PI) * tau_o) - erfc_theta_o); 
  double S_i = (1 - 0.5 * erfc_theta_i) / (1 + Lambda_i); 
  double S_o = (1 - 0.5 * erfc_theta_o) / (1 + Lambda_o); 
  double S = S_i * S_o;

  // Computing D and F (equations 78 to 83)  -- channel dependent terms
  double K_i = tan(theta_i) * erfc_theta_i;  // Equation 82
  double K_r = tan(theta_o) * erfc_theta_o;  // Equation 83

  double z_0_over_sigma_0 = solveForZ0(0.25 * sqrt(2./M_PI) * (K_i + K_r));    // Equation 81 -- solve for z0
  double sigma = sigma_0 * pow(1 + z_0_over_sigma_0 * z_0_over_sigma_0, -0.5); // Equation 80

  for (int channel = 0; channel < 3; channel++) {
  	  double f = 2 * sin(theta_h) * cos(theta_d); // RP: THE NEEXT TWO TERMS ARE CONSTANT PER LOOP ITERATION !!!
  	                                              // RP: Move them above ?!
  	  double v_xy2 = f * f;                                        // Equation 84
	  double F = fresnel4(n[channel], k[channel], cos(theta_d));   // Equation 75
	  double g = (2. * M_PI * sigma / lambda[channel]) * (cos(theta_i) + cos(theta_o)); 
	  g *= g;                                                      // Equation 79
	  double D = 0; 
	  for (int m = 1; m < 10; m++) {                               // Equation 78 -- infinite loop
	  	 // RP: factorial(m) * m  :  should be precomputed once for efficiency reason
	  	  D += (pow(g, m) * exp(-g)/(gsl_sf_fact(m) * m)) * exp(- 0.25 * v_xy2 * tau * tau/m);
		  if (isnan(D)) std::cout << "g = " << g << " m = " << m << std::endl;
	  }
	  D *= (0.5 * M_PI * tau / lambda[channel]) * (0.5 * M_PI * tau / lambda[channel]);
	  result[channel] = diffuse[channel] + F * D * S * G / (M_PI * cos(theta_i) * cos(theta_o)); // Equation 71
  }
}

void measureSpherical(double *initialInput, double *measurement, int dataSize, int numSamples, void *adata)
{
	// Levenberg-Marquadt for diffraction lobe, with Cook-Torrance lobe and Fresnel. 
	// 
	// Parameters: n[3], k[3]: index of refraction
	// alpha, b, p: shape distribution
	// sigma_s : std deviation of surface height
	double ior_phase[3]; 
	double ior_modulus[3]; 
	double n[3]; 
	double k[3]; 
	double diffuse[3]; 
	double sigma_0 = 1.0, tau = 1.0;
	int h = 0;
	double debugError = 0;

	// Constraint: n can be < 1, k can be null, but |n + ik| > 1.
	ior_modulus[0] = initialInput[h++];
	ior_modulus[1] = initialInput[h++];
	ior_modulus[2] = initialInput[h++];

	ior_phase[0] = initialInput[h++];
	ior_phase[1] = initialInput[h++];
	ior_phase[2] = initialInput[h++];

	n[0] = ior_modulus[0] * cos(ior_phase[0]);
	n[1] = ior_modulus[1] * cos(ior_phase[1]);
	n[2] = ior_modulus[2] * cos(ior_phase[2]);

	k[0] = ior_modulus[0] * sin(ior_phase[0]);
	k[1] = ior_modulus[1] * sin(ior_phase[1]);
	k[2] = ior_modulus[2] * sin(ior_phase[2]);

	sigma_0 = initialInput[h++];
	tau = initialInput[h++];

	diffuse[0] = initialInput[h++];
	diffuse[1] = initialInput[h++];
	diffuse[2] = initialInput[h++];
	

	double d_phi_o = M_PI /(2. * N);

	for(int i=0; i < numSamples; ++i)
		measurement[i] = 0.0;

	for (int kk = 0; kk < N_theta_i; kk++) {
		double theta_i = kk * d_theta_i;
		double phi_i = 0.0;
		for (int i = 0; i < N; i++) {
			double u_i = (double)i / N; // in [0,1]
			double theta_o = 0.5 * M_PI * u_i; 
			double d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
			for (int j = 0; j <= 2 * N; j++) {
				double phi_o = M_PI * j /(2. * N); // goes from M_PI to 0
				if ((theta_i < cutoff_theta) && (theta_o < cutoff_theta)) {
					if ((j == 0) || (j == 2* N)) d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N ;
					else d_omega = 2. * d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
					double mesured[3];
					double estimate[3];
					double theta_h, theta_d, phi_h, phi_d;
					MERLBRDF::lookup_brdf_val(brdfMesured, theta_i, phi_i, theta_o, phi_o, mesured[0], mesured[1], mesured[2], theta_h, phi_h, theta_d, phi_d);
					if (theta_d > cutoff_theta_d) continue;
					brdf_values(estimate, n, k, diffuse, sigma_0, tau, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o, phi_d);
					if (isnan(estimate[0])) std::cout << "# Isnan : " << theta_h << " " << theta_d << " " << theta_i << " " << theta_o << std::endl;
					// Debug, cause damn

					int ind = j + (2 * N + 1) * i + ((2 * N + 1) * N) * kk;
					assert(ind < skip); 
					double cosineTerm = 1.0;
					if (cosineFactor) cosineTerm = cos(theta_i);
					for (int channel = 0; channel < 3; channel++) {
						if (useLogNorm) {
							double Tp = log (1 + estimate[channel]);
							double Tm = log (1 + mesured[channel]);
							// Norm based on log
							measurement[ind + skip * channel] = d_omega * (fabs(Tm - Tp)); // LevMar is doing the squares and the summation
						} else {
							if (estimateOverMedian) cosineTerm *= compressiveWeight(estimate[channel]/median[channel]);
							else cosineTerm *= compressiveWeight(mesured[channel]/median[channel]);
							// Norm L2, sqrt(d_omega) since levmar is squaring it.
							measurement[ind + skip * channel] = sqrt(d_omega * cosineTerm) * 
								(fabs(estimate[channel] - mesured[channel])); // LevMar is doing the squares and the summation
						}
					}
				}
			} 
		}
	}
	// std::cerr << sqrt(debugError) << std::endl;
}

void measure1dimension(double *initialInput, double *measurement, int dataSize, int numSamples, void *adata)
{
	// Levenberg-Marquadt for diffraction lobe, with Cook-Torrance lobe and Fresnel. 
	// Fitting in 1D, over only theta_h
	// 
	// Parameters: n[3], k[3]: index of refraction
	// alpha, b, p: shape distribution
	// sigma_s : std deviation of surface height
	double ior_phase[3]; 
	double ior_modulus[3]; 
	double n[3]; 
	double k[3]; 
	double diffuse[3]; 
	double sigma_0 = 1.0, tau = 1.0;
	int h = 0;

	// Constraint: n can be < 1, k can be null, but |n + ik| > 1.
	ior_modulus[0] = initialInput[h++];
	ior_modulus[1] = initialInput[h++];
	ior_modulus[2] = initialInput[h++];

	ior_phase[0] = initialInput[h++];
	ior_phase[1] = initialInput[h++];
	ior_phase[2] = initialInput[h++];

	n[0] = ior_modulus[0] * cos(ior_phase[0]);
	n[1] = ior_modulus[1] * cos(ior_phase[1]);
	n[2] = ior_modulus[2] * cos(ior_phase[2]);

	k[0] = ior_modulus[0] * sin(ior_phase[0]);
	k[1] = ior_modulus[1] * sin(ior_phase[1]);
	k[2] = ior_modulus[2] * sin(ior_phase[2]);

	sigma_0 = initialInput[h++];
	tau = initialInput[h++];

	diffuse[0] = initialInput[h++];
	diffuse[1] = initialInput[h++];
	diffuse[2] = initialInput[h++];

	int skip = BRDF_SAMPLING_RES_THETA_H; 
	for(int i=0; i < numSamples; ++i)
		measurement[i] = 0.0;

	for (int j = 0; j <  BRDF_SAMPLING_RES_THETA_H ; j++) {
		double theta_h = j * j * M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H);
		if (theta_h >= cutoff_theta_i) continue; 
		double d_omega = sqrt(2 * j *  M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H));
		double theta_d = 0;
		double phi_h = 0.0;
		double phi_d = 0.0; // M_PI; // 0.5 * M_PI; 
		double theta_i, theta_o, phi_i, phi_o; 
		double mesured[3];
		double estimate[3];

		int ind = MERLBRDF::phi_diff_index(phi_d) +
			MERLBRDF::theta_diff_index(theta_d) * BRDF_SAMPLING_RES_PHI_D / 2 +
			MERLBRDF::theta_half_index(theta_h) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

		mesured[0] = brdfMesured[ind] * RED_SCALE;
		mesured[1] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		mesured[2] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;
		theta_i = theta_h; 
		theta_o = theta_h;
		theta_d = 0; 
		phi_o = 0;
		phi_i = 0;
		phi_d = 0;

		brdf_values(estimate, n, k, diffuse, sigma_0, tau, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o, phi_d);
		if (isnan(estimate[0])) std::cout << "# Isnan : " << theta_h << " " << theta_d << " " << theta_i << " " << theta_o << std::endl;
		ind = j ;
		assert(ind < skip); 
		for (int channel = 0; channel < 3; channel++) {
			if (ind + skip * channel >= numSamples) {
				std::cerr << "ind = " << ind << " skip = " << skip << " channel = " << channel << " numSmaples = " << numSamples << std::endl;
			}
			assert(ind + skip * channel < numSamples);
			double cosineTerm = 1.0;
			if (cosineFactor) cosineTerm = cos(theta_i);
			if (useLogNorm) {
				double Tp = log (1 + estimate[channel]);
				double Tm = log (1 + mesured[channel]);
				// Norm based on log
				measurement[ind + skip * channel] = d_omega * (fabs(Tm - Tp)); // LevMar is doing the squares and the summation
			} else {
				if (estimateOverMedian) cosineTerm *= compressiveWeight(estimate[channel]/median[channel]);
				else cosineTerm *= compressiveWeight(mesured[channel]/median[channel]);
				// Norm L2, sqrt(d_omega) since levmar is squaring it.
				measurement[ind + skip * channel] = d_omega * sqrt(cosineTerm) * 
					(fabs(estimate[channel] - mesured[channel])); // LevMar is doing the squares and the summation
			}
			// debugError += measurement[ind + skip * channel] * measurement[ind + skip * channel];
		}
	}
} 


void displaySolution(double initialInput[11]) 
{
	double ior_phase[3]; 
	double ior_modulus[3]; 
	double n[3]; 
	double k[3]; 
	double diffuse[3]; 
	double sigma_0 = 1.0, tau = 1.0;
	int h = 0;

	// Constraint: n can be < 1, k can be null, but |n + ik| > 1.
	ior_modulus[0] = initialInput[h++];
	ior_modulus[1] = initialInput[h++];
	ior_modulus[2] = initialInput[h++];

	ior_phase[0] = initialInput[h++];
	ior_phase[1] = initialInput[h++];
	ior_phase[2] = initialInput[h++];

	n[0] = ior_modulus[0] * cos(ior_phase[0]);
	n[1] = ior_modulus[1] * cos(ior_phase[1]);
	n[2] = ior_modulus[2] * cos(ior_phase[2]);

	k[0] = ior_modulus[0] * sin(ior_phase[0]);
	k[1] = ior_modulus[1] * sin(ior_phase[1]);
	k[2] = ior_modulus[2] * sin(ior_phase[2]);
	
	sigma_0 = initialInput[h++];
	tau = initialInput[h++];
	
	diffuse[0] = initialInput[h++];
	diffuse[1] = initialInput[h++];
	diffuse[2] = initialInput[h++];
	

	for (int channel = 0; channel < 3; channel++) {
		double color = fresnel4(n[channel], k[channel], 1.0); 
		std::cout << "# " << lambda[channel] << " " << n[channel] << " " << k[channel] << " " << color << " " << ior_modulus[channel] << " " << ior_phase[channel] << std::endl;
	}
	for (int channel = 0; channel < 3; channel++) {
		double color = fresnel4(n[channel], k[channel], 1.0); 
		std::cout << "# n[" << channel << "] = " << n[channel] << ";" << std::endl;
		std::cout << "# k[" << channel << "] = " << k[channel] << ";" << std::endl;
	}

	std::cout << "# { name  , {"
		<< n[0] << ", " 
		<< n[1] << ", " 
		<< n[2] << "}, {" 
		<< k[0] << ", " 
		<< k[1] << ", " 
		<< k[2] << "}, " 
		<< sigma_0 << ", " <<  tau << ", {" << diffuse[0] << ", " << diffuse[1] << ", " << diffuse[2] << "}";
	std::cout << "}," << std::endl;

	// Display solution
	// With He model, can't display back-scattering. Will display forward: theta_i = theta_o, phi_o = M_PI
	double mesured[3];
	double estimate[3];
	double reflection[3];
	double d_phi_o = M_PI /(2. * N);

	double theta_i = M_PI / 6.; 
	double phi_i = 0.0;
	for (int i = 0; i < N; i++) {
		double u_i = (double)i / N; // in [0,1]
		double theta_o = 0.5 * M_PI * u_i; 
		double d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
		for (int j = 0; j <= 2 * N; j++) {
			double phi_o = M_PI * j /(2. * N); // goes from M_PI to 0
			if ((theta_i < cutoff_theta) && (theta_o < cutoff_theta)) {
				if ((j == 0) || (j == 2* N)) d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N ;
				else d_omega = 2. * d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
				double mesured[3];
				double estimate[3];
				double theta_h, theta_d, phi_h, phi_d;
				MERLBRDF::lookup_brdf_val(brdfMesured, theta_i, phi_i, theta_o, phi_o, mesured[0], mesured[1], mesured[2], theta_h, phi_h, theta_d, phi_d);
				if (theta_d > cutoff_theta_d) continue;
				brdf_values(estimate, n, k, diffuse, sigma_0, tau, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o, phi_d);
				std::cout << theta_o << " " << phi_o << " ";
				for (int channel = 0; channel < 3; channel++) {
					std::cout << mesured[channel] << " " << estimate[channel] << " " << diffuse[channel] << " ";
				}
				std::cout << std::endl;
			}
		}
		std::cout << std::endl; 
	}
	std::cout << std::endl;
	std::cout << std::endl;
}


void readData(const char* filepath)
{	
	//	cosineFactor = false;
  // read brdf from file
  if (!MERLBRDF::read_brdf(filepath, brdfMesured)) 
  {
    fprintf(stderr, "Error reading %s\n", filepath);
    exit(1);
  }

  scaleReading[0] = RED_SCALE;
  scaleReading[1] = GREEN_SCALE;
  scaleReading[2] = BLUE_SCALE; 
  // Theoretical RGB values 
  //RP:  WHY ARE wavelength in micrometers ??
  lambda[0] = 0.645; 
  lambda[1] = 0.526; 
  lambda[2] = 0.444; 
  // Retiga 1300 values from the internet
  // works, too, but you have to remove the scaling in MERLBRDFRead.h
  // lambda[0] = 0.610; 
  // lambda[1] = 0.540; 
  // lambda[2] = 0.470; 
  //  From Bagher et al. TOG, 2016: weight samples according to relation to median BRDF value.
  //  Weighted by importance (wv and wi). This is just to get the median value
  computeMedian(brdfMesured, median); 

  const int maxDataSize = 12; // Max data size. 8 for fit without k, 11 for fits with k.
  const int maxNumSteps = 400; 
  const int numTestsTogether = 10;
  double errorsTogether[numTestsTogether];
  double solutionsTogether[numTestsTogether][maxDataSize];
  double initialData[maxDataSize];
  double lowerBounds[maxDataSize];
  double upperBounds[maxDataSize];
  double typicalValue[maxDataSize];
  double measurement[3 * BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_D];
  double *measurementTogether;
  double info[LM_INFO_SZ];
  double opts[5];
  bool showResults = true;
  // 1st test: all together, over entire range
  double errorMin;
  int kmin;
  int numErrorMinFound;

  cutoff_theta = 1.;
  cutoff_theta_i = 1.;
  N_theta_i = 20; // Number of tests in theta_i;
  N = 180;
  d_theta_i = cutoff_theta_i / N_theta_i; 
  skip = (2 * N + 1) * N * N_theta_i; 
  measurementTogether = new double[3 * skip];

  for (int k = 0; k < numTestsTogether; ) {
    for(int i=0; i< 3 * skip; ++i)
      measurementTogether[i] = 0.0;
    for (int i=0; i < LM_INFO_SZ; i++) info[i] = 0.0;
    opts[0] = 1.0;
    // These are easier options
    opts[1] = 1e-15;
    opts[2] = 1e-15;
    opts[3] = 1e-15;
	opts[4] = -1e-9;
    int dataSize = 0;

	// Modulus of index of refraction sqrt(n^2 + k^2). Theory says larger than 1.
	// Practice says larger than 1.5 for metals
    initialData[dataSize] = 1.1 + 2 * drand48(); 
    lowerBounds[dataSize] = 1.;
    upperBounds[dataSize] = 40.0;
    typicalValue[dataSize] = 3;
    dataSize += 1;
    initialData[dataSize] = initialData[0]; 
    lowerBounds[dataSize] = 1.;
    upperBounds[dataSize] = 40.0;
    typicalValue[dataSize] = 3;
    dataSize += 1;
    initialData[dataSize] = initialData[0]; 
    lowerBounds[dataSize] = 1.;
    upperBounds[dataSize] = 40.0;
    typicalValue[dataSize] = 3;
    dataSize += 1;
	// ior phase parameter
	initialData[dataSize] = 0.5 + 0.3 * drand48();        
	lowerBounds[dataSize] = 0.; 
	upperBounds[dataSize] = 0.5 * M_PI;
	typicalValue[dataSize] = 1.0;
	dataSize += 1;
	initialData[dataSize] = initialData[3]; 
	lowerBounds[dataSize] = 0.;
	upperBounds[dataSize] = 0.5 * M_PI;
	typicalValue[dataSize] = 1.0;
	dataSize += 1;
	initialData[dataSize] = initialData[3]; 
	lowerBounds[dataSize] = 0.;
	upperBounds[dataSize] = 0.5 * M_PI;
	typicalValue[dataSize] = 1.0;
	dataSize += 1;
	//
	// sigma_0, standard height deviation 
	initialData[dataSize] = 0.01 * drand48();
	lowerBounds[dataSize] = 1e-3; 
	upperBounds[dataSize] = 1000.0; 
	typicalValue[dataSize] = 0.1; // metals
	dataSize += 1;
	// tau, inverse of characteristic length
	initialData[dataSize] = 100 * drand48(); 
	lowerBounds[dataSize] = 1e-5; // at 1e-9, isnan for S.
	upperBounds[dataSize] = 1e10;
	typicalValue[dataSize] = 100;
	dataSize += 1;

	int reference = dataSize;
	initialData[dataSize] = 0.1 * drand48();        // diffuse
	lowerBounds[dataSize] = 0.0;
	upperBounds[dataSize] = 1.0;
	typicalValue[dataSize] = 0.1;
	dataSize += 1;
	initialData[dataSize] = initialData[reference]; // diffuse
	lowerBounds[dataSize] = 0.0;
	upperBounds[dataSize] = upperBounds[reference];
	typicalValue[dataSize] = typicalValue[reference];
	dataSize += 1;
	initialData[dataSize] = initialData[reference]; // diffuse
	lowerBounds[dataSize] = 0.0;
	upperBounds[dataSize] = upperBounds[reference];
	typicalValue[dataSize] = typicalValue[reference];
	dataSize += 1;

	for (int h = dataSize; h < maxDataSize; h++) initialData[h] = 0.0;

    printf("# Fitting all parameters together. k= %d\n# Initial data = ", k);
    for(int i=0; i < dataSize; ++i) {
      printf("%.7g ", initialData[i]);
    }
    printf("\n"); fflush(stdout);
    int ret;
// #define TWO_PASS
#ifdef TWO_PASS
	// LevMar, Box constraints, Analytic differentials
	ret = dlevmar_bc_dif(measure1dimension, initialData, measurementTogether, 
			dataSize, 3 * BRDF_SAMPLING_RES_THETA_H, 
			lowerBounds, upperBounds, typicalValue, 
			4000, opts, info, NULL, NULL, NULL);
	printf("# Levenberg-Marquardt 1D returned %d in %g iter, reason %g, error = %g, k= %d\n", ret, info[5], info[6], info[1], k);
	fflush(stdout);

	// Only increment k if 1st step converged and E < 1e-3
	// Otherwise, retry. More 1D tests, less useless 2D tests
	if ((info[6] <= 3) && (info[1] < 5.0 /*3e-3*/)) {
		for (int i = 0; i < maxDataSize; i++) solutionsTogether[k][i] =  initialData[i];
		if (showResults) displaySolution(initialData); 
		for (int h = 6; h < dataSize; h++) {
			// Typical value is now what we found in previous step
			if (initialData[h] > 0) typicalValue[h] = initialData[h];
			// Possible : update lower / upper bounds to allow only +- 30 %.
		}
#endif
		ret = dlevmar_bc_dif(measureSpherical, initialData, measurementTogether, 
	 			dataSize, 3 * skip, 
	 			lowerBounds, upperBounds, typicalValue, 
	 			20, opts, info, NULL, NULL, NULL);
		// ret = dlevmar_bc_dif(measureSphericalConstrained, initialData, measurementTogether, 
	 	// 		6, 3 * skip, 
	 	// 		lowerBounds, upperBounds, typicalValue, 
	 	// 		20, opts, info, NULL, NULL, initialData + 6);
	 	printf("# Levenberg-Marquardt 2D returned %d in %g iter, reason %g, error = %g, k= %d\n", ret, info[5], info[6], info[1], k);
	 	fflush(stdout);
	 	errorsTogether[k] = info[1]; // Error by fitting
	 	for (int i = 0; i < maxDataSize; i++) solutionsTogether[k][i] =  initialData[i];
	 	if (showResults) displaySolution(solutionsTogether[k]); 
		k += 1;
#ifdef TWO_PASS
	} else errorsTogether[k] = HUGE_VAL;
#endif
  }
  errorMin = HUGE_VAL;
  kmin = -1;
  numErrorMinFound = 0;
  for (int k = 0; k < numTestsTogether; k++) {
    if (errorsTogether[k] < errorMin) {
      errorMin = errorsTogether[k];
      kmin = k;
    }
  }
  if (kmin >= 0) {
    for (int k = 0; k < numTestsTogether; k++) {
      if (errorsTogether[k] < errorMin * 1.1) numErrorMinFound++;
    }
    std::cout << "# Best solution for Together = " << kmin << " error = " << errorMin << ", found " << numErrorMinFound << " times out of " << numTestsTogether << " (" << (float)numErrorMinFound*100./numTestsTogether << " %)"<< std::endl;
  }

  free(brdfMesured);
  delete [] measurementTogether;
}


int main(int argc, char* argv[])
{
  if(argc < 2) {
    std::cerr << "Usage: ./a.out brdf_name" << std::endl;
    exit(1);
  }
  const char *input_filename = argv[1];

  readData(input_filename);
  return 0;
}

