/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include "MERLBRDFRead.h"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <assert.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sf.h>
// This depends on the actual location of your levmar library: 
#include "../../../../../levmar-2.6/levmar.h"

// To compile:
// g++ -O3 -I /usr/local/include fit_conductors.cpp -lm -lstdc++ -L /usr/local/lib -lgsl ../../../../../levmar-2.6/liblevmar.a -llapack -lcblas -o conductors
// (depending on the actual location of GSL and levmar, adapt the above line)
//
// To run: ./conductors /path/to/MERL/reflectance.binary > reflectance.out
//
#include "common_functions.inl"

// Computes BRDF values but works on 3 channels at once. Saves computations.
void brdf_values(double result[3], double n[3],  double k[3], double b, double p, double alpha, double c, double sigma_s, double theta_h, double theta_d, double theta_i, double theta_o, double phi_i, double phi_o, double phi_d)
{
	result[0] = 0.0;
	result[1] = 0.0;
	result[2] = 0.0;
  if (theta_o >= 0.5 * M_PI) return;
  if (theta_i >= 0.5 * M_PI) return;
  
  double cookTorrance = 0;
  double diffraction = 0;
  double G2val = 1.0;
  double lowPassVal = 1.0;
  double multiple[3];
  multiple[0] = multiple[1] = multiple[2] = 0.0;


  if (useReflection) {
	  // Exponential micro-facet distribution -- channel independent part
	  G2val = G2(theta_i, theta_o, phi_i, phi_o, b, p) ;
	  if (!useGGX) lowPassVal = lowPassFilter(b, p, theta_i) * lowPassFilter(b, p, theta_o);
	  cookTorrance = 0.25 * D(theta_h, b, p);
	  cookTorrance *= lowPassVal * G2val / (cos(theta_i) * cos(theta_o));
	  // multipleScatter(multiple, n, k, b, p, theta_i, theta_o);
  } 

  // double f = 2 * sin(theta_h) * cos(theta_d); 

  
  for (int channel = 0; channel < 3; channel++) {
	  double F = fresnel4(n[channel], k[channel], cos(theta_d));
	  double l2 = lambda[channel] * lambda[channel];
	  double A_specular = 1.0, A_diffraction = 0.0;
	  double Q;
	  if (useDiffraction) {
		  // GGX/ABC for diffraction part:
		  double al = alpha * lambda[channel]; 
		  double u = cos(theta_d) / al;
		  double scale = 1.0; 
		  bool isDirac = true; 
		  double u_prime = convolutionResult(u, b, p, c, scale, isDirac); 
		  double f = 2 * sin(theta_h) * u_prime;
		  double f2 = f * f; 
		  double sigma_rel2_over_sigma_s2 = 1.0; 
		  if (isDirac) sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_i), c);
		  else sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_d), c);
		  // A_spec = portion of energy going into specular beam
		  double exp_term = 2. * M_PI * sigma_s * 2 * cos(theta_d);
		  exp_term *= exp_term;
		  // Multiplicative factor for specular reflection
		  A_specular = exp(- exp_term * sigma_rel2_over_sigma_s2/ l2);
		  if (isDirac) {
			  exp_term = 2. * M_PI * sigma_s * (cos(theta_i) + cos(theta_o));
			  exp_term *= exp_term;
			  A_diffraction = exp(- exp_term * sigma_rel2_over_sigma_s2 / l2);
			  Q = fresnelQ(n[channel], k[channel], theta_i, theta_o, phi_o);
			  scale = 1; // should be useless
		  } else {
			  // Multiplicative factor for specular reflection
			  // A_specular = 1 - exp_term * sigma_rel2_over_sigma_s2 / l2;
			  // if (A_specular < 0) A_specular = 0;
			  A_diffraction = A_specular;
			  Q = 2. * fresnel4(n[channel], k[channel], cos(theta_d)); // fresnelQ(n[channel], k[channel], theta_d, theta_d, M_PI);
			  // Q = fresnelQ(n[channel], k[channel], theta_i, theta_o, phi_o);
		  }
		  // GHS with ABC PSD
		  diffraction = ((1. - A_diffraction) * Q * scale * (c - 1.)/ (2. * M_PI * al * al)) * pow(1 + f2, - 0.5 * (c+1));
		  // re-normalization:
		  diffraction /= sigma_rel2_over_sigma_s2;
	  }
	  result[channel] = F * A_specular * cookTorrance + G2val * diffraction + multiple[channel];
  }
}

void measureSpherical(double *initialInput, double *measurement, int dataSize, int numSamples, void *adata)
{
	// Levenberg-Marquadt for diffraction lobe, with Cook-Torrance lobe and Fresnel. 
	// 
	// Parameters: n[3], k[3]: index of refraction
	// alpha, b, p: shape distribution
	// sigma_s : std deviation of surface height
	double ior_phase[3]; 
	double ior_modulus[3]; 
	double n[3]; 
	double k[3]; 
	double bdiff = 0.3, c = 2.0, b = 1.0, p = 1.0;
	double sigma_s = 0;
	int h = 0;

	// Constraint: n can be < 1, k can be null, but |n + ik| > 1.
	ior_modulus[0] = initialInput[h++];
	ior_modulus[1] = initialInput[h++];
	ior_modulus[2] = initialInput[h++];

	ior_phase[0] = initialInput[h++];
	ior_phase[1] = initialInput[h++];
	ior_phase[2] = initialInput[h++];

	n[0] = ior_modulus[0] * cos(ior_phase[0]);
	n[1] = ior_modulus[1] * cos(ior_phase[1]);
	n[2] = ior_modulus[2] * cos(ior_phase[2]);

	k[0] = ior_modulus[0] * sin(ior_phase[0]);
	k[1] = ior_modulus[1] * sin(ior_phase[1]);
	k[2] = ior_modulus[2] * sin(ior_phase[2]);

	if (useDiffraction) bdiff = initialInput[h++];
	if (useDiffraction) c = initialInput[h++];
	if (useReflection) {
		b = initialInput[h++];
		if (!useGGX) {
			p = initialInput[h++];
			b *= bMaxValue(p); 
		}
	}
	if (useDiffraction) {
		sigma_s = initialInput[h++]; 
	}

	double d_phi_o = M_PI /(2. * N);

	for(int i=0; i < numSamples; ++i)
		measurement[i] = 0.0;

	for (int kk = 0; kk < N_theta_i; kk++) {
		double theta_i = kk * d_theta_i;
		double phi_i = 0.0;
		for (int i = 0; i < N; i++) {
			double u_i = (double)i / N; // in [0,1]
			double theta_o = 0.5 * M_PI * u_i; 
			double d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
			for (int j = 0; j <= 2 * N; j++) {
				double phi_o = M_PI * j /(2. * N); // goes from M_PI to 0
				if ((theta_i < cutoff_theta) && (theta_o < cutoff_theta)) {
					if ((j == 0) || (j == 2* N)) d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N ;
					else d_omega = 2. * d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
					double mesured[3];
					double estimate[3];
					double theta_h, theta_d, phi_h, phi_d;
					MERLBRDF::lookup_brdf_val(brdfMesured, theta_i, phi_i, theta_o, phi_o, mesured[0], mesured[1], mesured[2], theta_h, phi_h, theta_d, phi_d);
					if (theta_d > cutoff_theta_d) continue;
					brdf_values(estimate, n, k, b, p, 1./bdiff, c, sigma_s, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o, phi_d);
					int ind = j + (2 * N + 1) * i + ((2 * N + 1) * N) * kk;
					assert(ind < skip); 
					double cosineTerm = 1.0;
					if (cosineFactor) cosineTerm = cos(theta_i);
					for (int channel = 0; channel < 3; channel++) {
						if (useLogNorm) {
							double Tp = log (1 + estimate[channel]);
							double Tm = log (1 + mesured[channel]);
							// Norm based on log
							measurement[ind + skip * channel] = d_omega * (fabs(Tm - Tp)); // LevMar is doing the squares and the summation
						} else {
							if (estimateOverMedian) cosineTerm *= compressiveWeight(estimate[channel]/median[channel]);
							else cosineTerm *= compressiveWeight(mesured[channel]/median[channel]);
							// Norm L2, sqrt(d_omega) since levmar is squaring it.
							measurement[ind + skip * channel] = sqrt(d_omega * cosineTerm) * 
								(fabs(estimate[channel] - mesured[channel])); // LevMar is doing the squares and the summation
						}
					}
				}
			} 
		}
	}
}

void measure1dimension(double *initialInput, double *measurement, int dataSize, int numSamples, void *adata)
{
	// Levenberg-Marquadt for diffraction lobe, with Cook-Torrance lobe and Fresnel. 
	// Fitting in 1D, over only theta_h
	// 
	// Parameters: n[3], k[3]: index of refraction
	// alpha, b, p: shape distribution
	// sigma_s : std deviation of surface height
	double ior_phase[3]; 
	double ior_modulus[3]; 
	double n[3]; 
	double k[3]; 
	double bdiff = 1.0, c = 2.0, b = 1.0, p = 1.0;
	double sigma_s = 0;
	int h = 0;

	// Constraint: n can be < 1, k can be null, but |n + ik| > 1.
	ior_modulus[0] = initialInput[h++];
	ior_modulus[1] = initialInput[h++];
	ior_modulus[2] = initialInput[h++];

	ior_phase[0] = initialInput[h++];
	ior_phase[1] = initialInput[h++];
	ior_phase[2] = initialInput[h++];

	n[0] = ior_modulus[0] * cos(ior_phase[0]);
	n[1] = ior_modulus[1] * cos(ior_phase[1]);
	n[2] = ior_modulus[2] * cos(ior_phase[2]);

	k[0] = ior_modulus[0] * sin(ior_phase[0]);
	k[1] = ior_modulus[1] * sin(ior_phase[1]);
	k[2] = ior_modulus[2] * sin(ior_phase[2]);

	if (useDiffraction) bdiff = initialInput[h++];
	if (useDiffraction) c = initialInput[h++];
	if (useReflection) {
		b = initialInput[h++];
		if (!useGGX) {
			p = initialInput[h++];
			b *= bMaxValue(p); 
		} 
	}
	if (useDiffraction) {
		sigma_s = initialInput[h++]; 
	}

	int skip = BRDF_SAMPLING_RES_THETA_H; 
	for(int i=0; i < numSamples; ++i)
		measurement[i] = 0.0;

	for (int j = 0; j <  BRDF_SAMPLING_RES_THETA_H ; j++) {
		double theta_h = j * j * M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H);
		if (theta_h >= cutoff_theta_i) continue; 
		double d_omega = sqrt(2 * j *  M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H));
		double theta_d = 0;
		double phi_h = 0.0;
		double phi_d = 0.0; // M_PI; // 0.5 * M_PI; 
		double theta_i, theta_o, phi_i, phi_o; 
		double mesured[3];
		double estimate[3];

		int ind = MERLBRDF::phi_diff_index(phi_d) +
			MERLBRDF::theta_diff_index(theta_d) * BRDF_SAMPLING_RES_PHI_D / 2 +
			MERLBRDF::theta_half_index(theta_h) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

		mesured[0] = brdfMesured[ind] * RED_SCALE;
		mesured[1] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		mesured[2] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;
		theta_i = theta_h; 
		theta_o = theta_h;
		theta_d = 0; 
		phi_o = 0;
		phi_i = 0;
		phi_d = 0;

		brdf_values(estimate, n, k, b, p, 1./bdiff, c, sigma_s, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o, phi_d);
		if (isnan(estimate[0])) std::cout << "# Isnan : " << theta_h << " " << theta_d << " " << theta_i << " " << theta_o << std::endl;
		ind = j ;
		assert(ind < skip); 
		double cosineTerm = 1.0;
		if (cosineFactor) cosineTerm = cos(theta_i);
		for (int channel = 0; channel < 3; channel++) {
			if (ind + skip * channel >= numSamples) {
				std::cerr << "ind = " << ind << " skip = " << skip << " channel = " << channel << " numSmaples = " << numSamples << std::endl;
			}
			assert(ind + skip * channel < numSamples);
			if (useLogNorm) {
				double Tp = log (1 + estimate[channel]);
				double Tm = log (1 + mesured[channel]);
				// Norm based on log
				measurement[ind + skip * channel] = d_omega * (fabs(Tm - Tp)); // LevMar is doing the squares and the summation
			} else {
				if (estimateOverMedian) cosineTerm *= compressiveWeight(estimate[channel]/median[channel]);
				else cosineTerm *= compressiveWeight(mesured[channel]/median[channel]);
				// Norm L2, sqrt(d_omega) since levmar is squaring it.
				measurement[ind + skip * channel] = d_omega * sqrt(cosineTerm) * 
					(fabs(estimate[channel] - mesured[channel])); // LevMar is doing the squares and the summation
			}
			// debugError += measurement[ind + skip * channel] * measurement[ind + skip * channel];
		}
	}
} 


void displaySolution(double initialInput[11]) 
{
	double ior_phase[3]; 
	double ior_modulus[3]; 
	double n[3]; 
	double k[3]; 
	double bdiff = 1.0, c = 2.0, b = 1.0, p = 0.0;
	double sigma_s = 0;
	int h = 0;

	// Constraint: n can be < 1, k can be null, but |n + ik| > 1.
	ior_modulus[0] = initialInput[h++];
	ior_modulus[1] = initialInput[h++];
	ior_modulus[2] = initialInput[h++];

	ior_phase[0] = initialInput[h++];
	ior_phase[1] = initialInput[h++];
	ior_phase[2] = initialInput[h++];

	n[0] = ior_modulus[0] * cos(ior_phase[0]);
	n[1] = ior_modulus[1] * cos(ior_phase[1]);
	n[2] = ior_modulus[2] * cos(ior_phase[2]);

	k[0] = ior_modulus[0] * sin(ior_phase[0]);
	k[1] = ior_modulus[1] * sin(ior_phase[1]);
	k[2] = ior_modulus[2] * sin(ior_phase[2]);
	
	if (useDiffraction) bdiff = initialInput[h++];
	if (useDiffraction) c = initialInput[h++];
	if (useReflection) {
		b = initialInput[h++];
		if (!useGGX) {
			p = initialInput[h++];
			b *= bMaxValue(p);
		}
	}
	if (useDiffraction) {
		sigma_s = initialInput[h++]; 
	}

	for (int channel = 0; channel < 3; channel++) {
		double color = fresnel4(n[channel], k[channel], 1.0); 
		std::cout << "# " << lambda[channel] << " " << n[channel] << " " << k[channel] << " " << color << " " << ior_modulus[channel] << " " << ior_phase[channel] << std::endl;
	}
	for (int channel = 0; channel < 3; channel++) {
		double color = fresnel4(n[channel], k[channel], 1.0); 
		std::cout << "# n[" << channel << "] = " << n[channel] << ";" << std::endl;
		std::cout << "# k[" << channel << "] = " << k[channel] << ";" << std::endl;
	}

	if (useReflection) std::cout << "#  beta = " << b << ";\n"; 
	if (!useGGX) std::cout << "#    p = " << p << ";" << std::endl;
	else std::cout << "# // GGX" << std::endl; 
	if (useDiffraction) std::cout << "# b = " << bdiff << ";\n# c = " << c << ";" << std::endl;
    if (useReflection) std::cout << "# // scaling factor for D = " << 0.25 * D(0, b, p) << std::endl;
	
	std::cout << "# sigma_s = " << sigma_s << ";" << std::endl;
	std::cout << "# { name  , {"
		<< n[0] << ", " 
		<< n[1] << ", " 
		<< n[2] << "}, {" 
		<< k[0] << ", " 
		<< k[1] << ", " 
		<< k[2] << "}, " 
		<< b << ", " <<  p << ", " << sigma_s << ", " << 1./bdiff << ", " << c << ", conductor, ";
	std::cout << "}," << std::endl;

	// Display solution
	//
	double mesured[3];
	double estimate[3];
	double reflection[3];

	for (int j = 0; j <  BRDF_SAMPLING_RES_THETA_H ; j++) {
		double theta_h =  j * j * M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H);
		double theta_d = 0;
		double phi_h = 0.0;
		double phi_d = M_PI; // 0.5 * M_PI; 
		double theta_i, theta_o, phi_i, phi_o; 
		MERLBRDF::half_diff_coords_to_std_coords(theta_h, phi_h, theta_d, phi_d,
				theta_i, phi_i, theta_o, phi_o);

		std::cout << theta_h << " " ;

		int ind = MERLBRDF::phi_diff_index(phi_d) +
			MERLBRDF::theta_diff_index(theta_d) * BRDF_SAMPLING_RES_PHI_D / 2 +
			MERLBRDF::theta_half_index(theta_h) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

		mesured[0] = brdfMesured[ind] * RED_SCALE;
		mesured[1] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		mesured[2] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;
		brdf_values(estimate, n, k, b, p, 1./bdiff, c, sigma_s, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o, phi_d);
		// Computing BRDF values here for debugging
		double cookTorrance = 0;
		double diffraction = 0;
		double G2val = 1.0;
		double multiple[3];
		double lowPassVal = 0;
		double result;
		multiple[0] = multiple[1] = multiple[2] = 0.0;

		if (useReflection) {
			// Exponential micro-facet distribution -- channel independent part
			G2val = G2(theta_i, theta_o, phi_i, phi_o, b, p) ;
			if (!useGGX) lowPassVal = lowPassFilter(b, p, theta_i) * lowPassFilter(b, p, theta_o);
			cookTorrance = 0.25 * D(theta_h, b, p);
			cookTorrance *= lowPassVal * G2val / (cos(theta_i) * cos(theta_o));
			// multipleScatter(multiple, n, k, b, p, theta_i, theta_o);
		}
		//
		// double f = 2 * sin(theta_h) * cos(theta_d); 
		for (int channel = 0; channel < 3; channel++) {
			double F = fresnel4(n[channel], k[channel], cos(theta_d));
			double l2 = lambda[channel] * lambda[channel];
			// Multiplicative factor for specular reflection
			// A = portion of energy going into specular beam
			double A_specular = 1.0, A_diffraction = 0.0;
			double Q = 0;
			double sigma_rel2_over_sigma_s2 = 1.0; 
			if (useDiffraction) {
				// GGX/ABC for diffraction part:
				double al = lambda[channel] / bdiff; 
				double u = cos(theta_d) / al;
				double scale = 1.0; 
				bool isDirac = true; 
				double u_prime = convolutionResult(u, b, p, c, scale, isDirac); 
				double f = 2 * sin(theta_h) * u_prime;
				double f2 = f * f;
				if (isDirac) sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_i), c);
				else sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_d), c);
				// A_spec = portion of energy going into specular beam
				double exp_term = 2. * M_PI * sigma_s * 2 * cos(theta_d);
				exp_term *= exp_term;
				A_specular = exp(- exp_term * sigma_rel2_over_sigma_s2/ l2);
				// A_diff = portion of energy going into diffraction beam
				if (isDirac) {
					exp_term = 2. * M_PI * sigma_s * (cos(theta_i) + cos(theta_o));
					exp_term *= exp_term;
					A_diffraction = exp(- exp_term * sigma_rel2_over_sigma_s2 / l2);
					Q = fresnelQ(n[channel], k[channel], theta_i, theta_o, phi_o);
					scale = 1; // should be useless
				} else {
					// Multiplicative factor for specular reflection
					// A_specular = 1 - exp_term * sigma_rel2_over_sigma_s2 / l2;
					// if (A_specular < 0) A_specular = 0;
					A_diffraction = A_specular;
					Q = 2. * fresnel4(n[channel], k[channel], cos(theta_d)); // fresnelQ(n[channel], k[channel], theta_d, theta_d, M_PI);
				}
				// GHS with ABC PSD
				diffraction = ((1. - A_diffraction) * Q * scale * (c - 1.)/ (2. * M_PI * al * al)) * pow(1 + f2, - 0.5 * (c+1));
				// re-normalization:
				diffraction /= sigma_rel2_over_sigma_s2;
				
			}
			std::cout << mesured[channel] << " " << estimate[channel] << " " <<
			  F * A_specular * cookTorrance << " " << diffraction * G2val << " " << multiple[channel] << " ";
		}
		std::cout << std::endl;
	}
	std::cout << std::endl;
	std::cout << std::endl;
}


void readData(const char* filepath)
{	
	//	cosineFactor = false;
  // read brdf from file
  if (!MERLBRDF::read_brdf(filepath, brdfMesured)) 
  {
    fprintf(stderr, "Error reading %s\n", filepath);
    exit(1);
  }

  scaleReading[0] = RED_SCALE;
  scaleReading[1] = GREEN_SCALE;
  scaleReading[2] = BLUE_SCALE; 
  // Theoretical RGB values 
  lambda[0] = 0.645; 
  lambda[1] = 0.526; 
  lambda[2] = 0.444; 
  // Retiga 1300 values from the internet
  // works, too, but you have to remove the scaling in MERLBRDFRead.h
  // lambda[0] = 0.610; 
  // lambda[1] = 0.540; 
  // lambda[2] = 0.470; 
  //  From Bagher et al. TOG, 2016: weight samples according to relation to median BRDF value.
  //  Weighted by importance (wv and wi). This is just to get the median value
  computeMedian(brdfMesured, median); 


  const int maxDataSize = 12; // Max data size. 8 for fit without k, 11 for fits with k.
  const int maxNumSteps = 400; 
  const int numTestsTogether = 10;
  double errorsTogether[numTestsTogether];
  double solutionsTogether[numTestsTogether][maxDataSize];
  double initialData[maxDataSize];
  double lowerBounds[maxDataSize];
  double upperBounds[maxDataSize];
  double typicalValue[maxDataSize];
  double measurement[3 * BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_D];
  double *measurementTogether;
  double info[LM_INFO_SZ];
  double opts[5];
  bool showResults = true;
  // 1st test: all together, over entire range
  double errorMin;
  int kmin;
  int numErrorMinFound;

  N_theta_i = 20; // Number of tests in theta_i;
  N = 180;
  d_theta_i = cutoff_theta_i / N_theta_i; 
  skip = (2 * N + 1) * N * N_theta_i; 
  measurementTogether = new double[3 * skip];

  for (int k = 0; k < numTestsTogether; ) {
    for(int i=0; i< 3 * skip; ++i)
      measurementTogether[i] = 0.0;
    for (int i=0; i < LM_INFO_SZ; i++) info[i] = 0.0;
    opts[0] = 1.0;
    // These are easier options
    opts[1] = 1e-15;
    opts[2] = 1e-15;
    opts[3] = 1e-15;
	opts[4] = -1e-9;
    int dataSize = 0;

	// Modulus of index of refraction sqrt(n^2 + k^2). Theory says larger than 1.
	// Practice says larger than 1.5 for metals
    initialData[dataSize] = 1.1 + 2 * drand48(); 
    lowerBounds[dataSize] = 1.1;
    upperBounds[dataSize] = 40.0;
    typicalValue[dataSize] = 3;
    dataSize += 1;
    initialData[dataSize] = initialData[0]; 
    lowerBounds[dataSize] = 1.1;
    upperBounds[dataSize] = 40.0;
    typicalValue[dataSize] = 3;
    dataSize += 1;
    initialData[dataSize] = initialData[0]; 
    lowerBounds[dataSize] = 1.1;
    upperBounds[dataSize] = 40.0;
    typicalValue[dataSize] = 3;
    dataSize += 1;
	// ior phase parameter
	int reference = dataSize;
	initialData[dataSize] = 0.5 + 0.3 * drand48();        
	lowerBounds[dataSize] = 0.; 
	upperBounds[dataSize] = 0.5 * M_PI;
	typicalValue[dataSize] = 1.0;
	dataSize += 1;
	initialData[dataSize] = initialData[3]; 
	lowerBounds[dataSize] = 0.;
	upperBounds[dataSize] = 0.5 * M_PI;
	typicalValue[dataSize] = 1.0;
	dataSize += 1;
	initialData[dataSize] = initialData[3]; 
	lowerBounds[dataSize] = 0.;
	upperBounds[dataSize] = 0.5 * M_PI;
	typicalValue[dataSize] = 1.0;
	dataSize += 1;
	//
	// shape parameter values:
    if (useDiffraction) {
    	// characteristic length b, in micro-meters. 
    	// larger than wavelength (0.3), smaller than 10 mum (10)
    	initialData[dataSize] = 0.2 + 3 * drand48(); 
    	lowerBounds[dataSize] = 0.2; // ???
    	upperBounds[dataSize] = 200; // ??? 
    	typicalValue[dataSize] = 5;
    	dataSize += 1;
	}
    if (useDiffraction) {
    	initialData[dataSize] = 1.3 + drand48(); // c 
    	lowerBounds[dataSize] = 1.1; // 1 is mandatory
    	upperBounds[dataSize] = 10.0; // ??? 
    	typicalValue[dataSize] = 1.4;
    	dataSize += 1;
	}
    if (useReflection) { 
#define Shiny
#ifdef Shiny
		// b, average slope. Between 0 and 1 (relative to bMax)
		initialData[dataSize] = 0.01 + 0.01 * drand48();      // metals
    	lowerBounds[dataSize] = 1e-6; 
    	upperBounds[dataSize] = 1.0; 
    	typicalValue[dataSize] = 1e-5; // metals
    	dataSize += 1;
		if (!useGGX) {
			initialData[dataSize] = 0.15 + 0.2 * drand48(); // p
			lowerBounds[dataSize] = 0.15;  
			upperBounds[dataSize] = 5.0;
			typicalValue[dataSize] = 0.3;
			dataSize += 1;
		}
#else
		initialData[dataSize] = 0.2 + 0.2 * drand48();     
    	lowerBounds[dataSize] = 0.01; 
    	upperBounds[dataSize] = 9.0; 
    	typicalValue[dataSize] = 0.4;
    	dataSize += 1;
		if (!useGGX) { 
			initialData[dataSize] = 0.8 + 0.8 * drand48(); // p 
			lowerBounds[dataSize] = 0.51;  
			upperBounds[dataSize] = 4.0;
			typicalValue[dataSize] = 0.9;
			dataSize += 1;
		}
#endif
	}
// #undef Shiny
#ifdef Shiny
  	double sigmaList[] = {0.03, 0.03, 0.01, 0.01, 0.001};
#else
  	 double sigmaList[] = {1.0, 1.0, 0.3, 0.3, 0.5};
	 // double sigmaList[] = {0.03, 0.03, 0.01, 0.01, 0.001};
#endif
	if (useDiffraction) {
		int step = (k / 2) % 5; // 0, 0, 1, 1, 2, 2, ..., 4
		double defaultValue = sigmaList[step];
    	initialData[dataSize] = defaultValue + defaultValue * drand48(); // sigma_s
    	lowerBounds[dataSize] = 0.001 * defaultValue;
    	upperBounds[dataSize] = 1.; // but anything above 0.1 is useless
    	typicalValue[dataSize] = defaultValue ;
    	dataSize += 1;
    }

	for (int h = dataSize; h < maxDataSize; h++) initialData[h] = 0.0;

    printf("# Fitting all parameters together. k= %d\n# Initial data = ", k);
    for(int i=0; i < dataSize; ++i) {
      printf("%.7g ", initialData[i]);
    }
    printf("\n"); fflush(stdout);
    int ret;
#define TWO_PASS
#ifdef TWO_PASS
	// LevMar, Box constraints, Analytic differentials
	ret = dlevmar_bc_dif(measure1dimension, initialData, measurementTogether, 
			dataSize, 3 * BRDF_SAMPLING_RES_THETA_H, 
			lowerBounds, upperBounds, typicalValue, 
			4000, opts, info, NULL, NULL, NULL);
	printf("# Levenberg-Marquardt 1D returned %d in %g iter, reason %g, error = %g, k= %d\n", ret, info[5], info[6], info[1], k);
	fflush(stdout);

	// Only increment k if 1st step converged and E < 1e-3
	// Otherwise, retry. More 1D tests, less useless 2D tests
	if ((info[6] <= 3) && (info[1] < 3e-3)) {
		for (int i = 0; i < maxDataSize; i++) solutionsTogether[k][i] =  initialData[i];
		if (showResults) displaySolution(initialData); 
		for (int h = 6; h < dataSize; h++) {
			// Typical value is now what we found in previous step
			if (initialData[h] > 0) typicalValue[h] = initialData[h];
			// Possible : update lower / upper bounds to allow only +- 30 %.
		}
#endif
		ret = dlevmar_bc_dif(measureSpherical, initialData, measurementTogether, 
	 			dataSize, 3 * skip, 
	 			lowerBounds, upperBounds, typicalValue, 
	 			20, opts, info, NULL, NULL, NULL);
	 	printf("# Levenberg-Marquardt 2D returned %d in %g iter, reason %g, error = %g, k= %d\n", ret, info[5], info[6], info[1], k);
	 	fflush(stdout);
	 	errorsTogether[k] = info[1]; // Error by fitting
	 	for (int i = 0; i < maxDataSize; i++) solutionsTogether[k][i] =  initialData[i];
	 	if (showResults) displaySolution(solutionsTogether[k]); 
		k += 1;
#ifdef TWO_PASS
	} else errorsTogether[k] = HUGE_VAL;
#endif
  }
  errorMin = HUGE_VAL;
  kmin = -1;
  numErrorMinFound = 0;
  for (int k = 0; k < numTestsTogether; k++) {
    if (errorsTogether[k] < errorMin) {
      errorMin = errorsTogether[k];
      kmin = k;
    }
  }
  if (kmin >= 0) {
    for (int k = 0; k < numTestsTogether; k++) {
      if (errorsTogether[k] < errorMin * 1.1) numErrorMinFound++;
    }
    std::cout << "# Best solution for Together = " << kmin << " error = " << errorMin << ", found " << numErrorMinFound << " times out of " << numTestsTogether << " (" << (float)numErrorMinFound*100./numTestsTogether << " %)"<< std::endl;
  }

  free(brdfMesured);
  delete [] measurementTogether;
}


int main(int argc, char* argv[])
{
  if(argc < 2) {
    std::cerr << "Usage: ./a.out brdf_name" << std::endl;
    exit(1);
  }
  const char *input_filename = argv[1];

  readData(input_filename);
  return 0;
}

