/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include "MERLBRDFRead.h"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <assert.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_sf.h>
// This depends on the actual location of your levmar library: 
#include "../../../../../levmar-2.6/levmar.h"

// To compile:
// g++ -O3 -I /usr/local/include fit_plastic.cpp -lm -lstdc++ -L /usr/local/lib -lgsl ../../../../../levmar-2.6/liblevmar.a -llapack -lcblas -o plastic
// (depending on the actual location of GSL and levmar, adapt the above line)
//
// To run: ./plastic /path/to/MERL/reflectance.binary > reflectance.out
//
#include "common_functions.inl"

// Computes BRDF values but works on 3 channels at once. Saves computations.
void brdf_values(double result[3], double ior[3], double b, double p, double alpha, double c, double sigma_s, double albedo[3], double theta_h, double theta_d, double theta_i, double theta_o, double phi_i, double phi_o)
{
	result[0] = 0.0;
	result[1] = 0.0;
	result[2] = 0.0;
	if (theta_o >= 0.5 * M_PI) return;
	if (theta_i >= 0.5 * M_PI) return;

	double cookTorrance = 0;
	double diffraction = 0;
	double G2val = 1.0;
	double lowPassVal = 1.0;
	double multiple[3];
	multiple[0] = multiple[1] = multiple[2] = 0.0;

	if (useReflection) {
		// Exponential micro-facet distribution -- channel independent part
		G2val = G2(theta_i, theta_o, phi_i, phi_o, b, p) ;
		lowPassVal = lowPassFilter(b, p, theta_i) * lowPassFilter(b, p, theta_o);
		cookTorrance = 0.25 * D(theta_h, b, p);
		cookTorrance *= lowPassVal * G2val / (cos(theta_i) * cos(theta_o));
		// multipleScatter(multiple, ior, b, p, theta_i, theta_o);
	} 
	double f = 2 * sin(theta_h) * cos(theta_d); 

	for (int channel = 0; channel < 3; channel++) {
		double F = fresnel4(ior[channel], 0.0, cos(theta_d));
		double l2 = lambda[channel] * lambda[channel];
		// Multiplicative factor for specular reflection
		// A = portion of energy going into specular beam
		double A_specular = 1.0, A_diffraction = 0.0;
		double Q;
		if (useDiffraction) {
			double al = alpha * lambda[channel]; 
			double u = cos(theta_d) / al;
			double scale = 1.0; 
			bool isDirac = true; 
			double u_prime = convolutionResult(u, b, p, c, scale, isDirac); 
			double f = 2 * sin(theta_h) * u_prime;
			double f2 = f * f; 
			double sigma_rel2_over_sigma_s2 = 1.0; 
			if (isDirac) sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_i), c);
			else  sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_d), c);
			// A_spec = portion of energy going into specular beam
			double exp_term = 2. * M_PI * sigma_s * 2 * cos(theta_d);
			exp_term *= exp_term;
			// Multiplicative factor for specular reflection
			A_specular = exp(- exp_term * sigma_rel2_over_sigma_s2/ l2);
			if (isDirac) {
				exp_term = 2. * M_PI * sigma_s * (cos(theta_i) + cos(theta_o));
				exp_term *= exp_term;
				A_diffraction = exp(- exp_term * sigma_rel2_over_sigma_s2 / l2);
				Q = fresnelQ(ior[channel], 0.0, theta_i, theta_o, phi_o);
				scale = 1; // should be useless
			} else {
				// Multiplicative factor for specular reflection
				// A_specular = 1 - exp_term * sigma_rel2_over_sigma_s2 / l2;
				// if (A_specular < 0) A_specular = 0;
				A_diffraction = A_specular;
				Q = fresnelQ(ior[channel], 0.0, theta_d, theta_d, M_PI);
				// Q = fresnelQ(ior[channel], k[channel], theta_i, theta_o, phi_o);
			}
			// GHS with ABC PSD
			diffraction = ((1. - A_diffraction) * Q * scale * (c - 1.)/ (2. * M_PI * al * al)) * pow(1 + f2, - 0.5 * (c+1));
			// re-normalization:
			diffraction /= sigma_rel2_over_sigma_s2;
		}
		// diffuse layer underneath, formula from Mitsuba plastic.cpp, with nonlinear = true:
		double F_dr_internal = Fdr(1.0/ior[channel]);
		// Single scattering BRDF 
		double cos_theta_o_inside = sqrt(1. - (sin(theta_o)/ior[channel]) *  (sin(theta_o)/ior[channel]));
		// double cos_theta_i_inside = sqrt(1. - (sin(theta_i)/ior[channel]) *  (sin(theta_i)/ior[channel]));
		double diffuse = albedo[channel] * (1. - fresnel4(ior[channel], 0.0, cos(theta_i))) * (1. - fresnel4(1./ior[channel], 0.0, cos_theta_o_inside)) / ( M_PI /* ior[channel] * ior[channel]*/ * (1.0 - albedo[channel] * F_dr_internal)); 
		// For future reference: factor 1/pi. Make sure is present in Mitsuba.
		result[channel] = (F * A_specular * cookTorrance + G2val * diffraction + multiple[channel] + diffuse);
	}
}

void measureSpherical(double *initialInput, double *measurement, int numSamples, int n, void *adata)
{
	// Levenberg-Marquadt for diffraction lobe, with Cook-Torrance lobe and Fresnel. 
	// Test with theta_i from 0 to 90, theta_o, phi_o importance sampled around reflected direction.
	// 
	// Parameters: n[3], k[3]: index of refraction
	// alpha, b, p: shape distribution
	double ior[3]; 
	double albedo[3]; 
	double bdiff = 0.3, c = 2.0, b = 1.0, p = 1.0;
	double sigma_s = 0;
	int h = 0;
	double g = 0;

	ior[0] = initialInput[h++];
	ior[1] = initialInput[h++];
	ior[2] = initialInput[h++];

	if (useDiffraction) bdiff = initialInput[h++];
	if (useDiffraction) c = initialInput[h++];
	if (useReflection) {
		b = initialInput[h++];
		if (!useGGX) {
			p = initialInput[h++];
			b *= bMaxValue(p); 
		}
	}
	if (useDiffraction) {
		sigma_s = initialInput[h++];
	}

	albedo[0] = initialInput[h++];
	albedo[1] = initialInput[h++];
	albedo[2] = initialInput[h++];

	double d_phi_o = M_PI /(2. * N);

	for (int i=0; i < numSamples; ++i)
		measurement[i] = 0.0;

	for (int kk = 0; kk < N_theta_i; kk++) {
		double theta_i = kk * d_theta_i;
		double phi_i = 0.0;
		for (int i = 0; i < N; i++) {
			double u_i = (double)i / N; // in [0,1]
			double theta_o = 0.5 * M_PI * u_i; 
			double d_omega = d_phi_o * sin(theta_o) * 0.5 * M_PI / N;
			for (int j = 0; j < 2 * N; j++) {
				double phi_o = M_PI * j /(2. * N); // goes from 0 to M_PI
				if ((theta_i < cutoff_theta) && (theta_o < cutoff_theta)) {
					double mesured[3];
					double estimate[3];
					double theta_h, theta_d, phi_h, phi_d;
					MERLBRDF::lookup_brdf_val(brdfMesured, theta_i, phi_i, theta_o, phi_o, mesured[0], mesured[1], mesured[2], theta_h, phi_h, theta_d, phi_d);
					brdf_values(estimate, ior, b, p, 1./bdiff, c, sigma_s, albedo, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o);
					int ind = j + (2 * N) * i + (2 * N * N) * kk;
					assert(ind < skip); 
					double cosineTerm = 1.0;
					if (cosineFactor) cosineTerm = cos(theta_i);
					for (int channel = 0; channel < 3; channel++) {
						if (useLogNorm) {
							double Tp = log (1 + estimate[channel]);
							double Tm = log (1 + mesured[channel]);
							// Norm based on log
							measurement[ind + skip * channel] = d_omega * (fabs(Tm - Tp)); // LevMar is doing the squares and the summation
						} else {
							if (estimateOverMedian) cosineTerm *= compressiveWeight(estimate[channel]/median[channel]);
							else cosineTerm *= compressiveWeight(mesured[channel]/median[channel]);
							// Norm L2, sqrt(d_omega) since levmar is squaring it.
							measurement[ind + skip * channel] = sqrt(d_omega * cosineTerm) * 
								(fabs(estimate[channel] - mesured[channel])); // LevMar is doing the squares and the summation
						}
					}
				}
			} 
		}
	}
}

void measure1dimension(double *initialInput, double *measurement, int dataSize, int n, void *adata)
{
	// Levenberg-Marquadt for diffraction lobe, with Cook-Torrance lobe and Fresnel. 
	// Test with theta_i from 0 to 90, theta_o, phi_o importance sampled around reflected direction.
	// Fitting in 1D, over only theta_h
	// 
	// Parameters: n[3], k[3]: index of refraction
	// alpha, b, p: shape distribution
	double ior[3]; 
	double albedo[3]; 
	double B, C = 0;
	double bdiff = 1.0, c = 2.0, b = 1.0, p = 1.0;
	double sigma_s = 0;
	int h = 0;

	ior[0] = initialInput[h++];
	ior[1] = initialInput[h++];
	ior[2] = initialInput[h++];

	if (useDiffraction) bdiff = initialInput[h++];
	if (useDiffraction) c = initialInput[h++];
	if (useReflection) {
		b = initialInput[h++];
		if (!useGGX) {
			p = initialInput[h++];
			b *= bMaxValue(p); 
		}
	}
	if (useDiffraction) {
		sigma_s = initialInput[h++]; 
	}
	
	albedo[0] = initialInput[h++];
	albedo[1] = initialInput[h++];
	albedo[2] = initialInput[h++];

	double deviation[3];
	double result[3];
	double d_phi_d = M_PI /(2. * N);

	int skip = BRDF_SAMPLING_RES_THETA_H; 
	for(int i=0; i < n; ++i)
		measurement[i] = 0.0;
	for (int j = 0; j <  BRDF_SAMPLING_RES_THETA_H ; j++) {
		double theta_h = j * j * M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H);
		if (theta_h >= cutoff_theta_i) continue; 
		double d_omega = sqrt(2 * j *  M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H));
		double theta_d = 0;
		double phi_h = 0.0;
		double phi_d = 0.0; // M_PI; // 0.5 * M_PI; 
		double theta_i, theta_o, phi_i, phi_o; 
		double mesured[3];
		double estimate[3];

		int ind = MERLBRDF::phi_diff_index(phi_d) +
			MERLBRDF::theta_diff_index(theta_d) * BRDF_SAMPLING_RES_PHI_D / 2 +
			MERLBRDF::theta_half_index(theta_h) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

		mesured[0] = brdfMesured[ind] * RED_SCALE;
		mesured[1] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		mesured[2] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;
		theta_i = theta_h; 
		theta_o = theta_h;
		theta_d = 0; 
		phi_o = 0;
		phi_i = 0;

		brdf_values(estimate, ior, b, p, 1./bdiff, c, sigma_s, albedo, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o);
		ind = j ;
		assert(ind < skip); 
        double cosineTerm = 1.0;
        if (cosineFactor) cosineTerm = cos(theta_i);
		for (int channel = 0; channel < 3; channel++) {
			if (useLogNorm) {
				double Tp = log (1 + cos(theta_i) * estimate[channel]);
				double Tm = log (1 + cos(theta_i) * mesured[channel]);
				// Norm based on log
				measurement[ind + skip * channel] = d_omega * (fabs(Tm - Tp)); // LevMar is doing the squares and the summation
			} else {
				if (estimateOverMedian) cosineTerm *= compressiveWeight(estimate[channel]/median[channel]);
				else cosineTerm *= compressiveWeight(mesured[channel]/median[channel]);
				// Norm L2, sqrt(d_omega) since levmar is squaring it.
				measurement[ind + skip * channel] = d_omega * sqrt(cosineTerm) * 
					(fabs(estimate[channel] - mesured[channel])); // LevMar is doing the squares and the summation
			}
		}
	}
} 



void displaySolution(double initialInput[11]) 
{
	double ior[3]; 
	double albedo[3]; 
	double bdiff = 1.0, c = 2.0, b = 1.0, p = 1.0;
	double B, C = 0.0;
	double sigma_s = 0;
	int h = 0;


	ior[0] = initialInput[h++];
	ior[1] = initialInput[h++];
	ior[2] = initialInput[h++];
	
	if (useDiffraction) bdiff = initialInput[h++];
	if (useDiffraction) c = initialInput[h++];
	if (useReflection) {
		b = initialInput[h++];
		if (!useGGX) {
			p = initialInput[h++];
			b *= bMaxValue(p); 
		}
	}
	if (useDiffraction) {
		sigma_s = initialInput[h++]; 
	}

	albedo[0] = initialInput[h++];
	albedo[1] = initialInput[h++];
	albedo[2] = initialInput[h++];

	for (int channel = 0; channel < 3; channel++) {
		double color = fresnel4(ior[channel], 0.0, 1.0); 
		std::cout << "# " << lambda[channel] << " " << ior[channel] << " " << color << std::endl;
	}
	std::cout << "#  beta = " << b << ";\n#    p = " << p << ";" << std::endl;
	if (useDiffraction) std::cout << "# b = " << bdiff << ";\n# c = " << c << ";" << std::endl;
	std::cout << "# sigma_s = " << sigma_s << ";" << std::endl;
	std::cout << "# albedo = " << albedo[0] << " " << albedo[1] << " " << albedo[2] << std::endl;
  	std::cout << "# { name  , {" << ior[0] << ", " << ior[1] << ", " << ior[2] << "}, {" << albedo[0] << ", " << albedo[1] << ", " << albedo[2] << "}, " << b << ", " <<  p << ", " << sigma_s << ", " << 1./bdiff << ", " << c << ", plastic, "; 
	std::cout << "}," << std::endl;

	// Display solution
	//
	double mesured[3];
	double estimate[3];
	double reflection[3];

	for (int j = 0; j <  BRDF_SAMPLING_RES_THETA_H ; j++) {
		double theta_h =  j * j * M_PI * 0.5 / (BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_H);
		double theta_d = 0;
		double phi_h = 0.0;
		double phi_d = 0.0; // M_PI; // 0.5 * M_PI; 
		double theta_i, theta_o, phi_i, phi_o; 
		MERLBRDF::half_diff_coords_to_std_coords(theta_h, phi_h, theta_d, phi_d,
				theta_i, phi_i, theta_o, phi_o);

		std::cout << theta_h << " " ;

		int ind = MERLBRDF::phi_diff_index(phi_d) +
			MERLBRDF::theta_diff_index(theta_d) * BRDF_SAMPLING_RES_PHI_D / 2 +
			MERLBRDF::theta_half_index(theta_h) * BRDF_SAMPLING_RES_PHI_D / 2 * BRDF_SAMPLING_RES_THETA_D;

		mesured[0] = brdfMesured[ind] * RED_SCALE;
		mesured[1] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2] * GREEN_SCALE;
		mesured[2] = brdfMesured[ind + BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D] * BLUE_SCALE;
		brdf_values(estimate, ior, b, p, 1./bdiff, c, sigma_s, albedo, theta_h, theta_d, theta_i, theta_o, phi_i, phi_o);
		// Computing BRDF values here for debugging
		double cookTorrance = 0;
		double diffraction = 0;
		double G2val = 1.0;
		// Exponential micro-facet distribution -- channel independent part
		
		G2val = G2(theta_i, theta_o, phi_i, phi_o, b, p) ;
		cookTorrance = 0.25 * D(theta_h, b, p);
		cookTorrance *= G2val / (cos(theta_i) * cos(theta_o));
		double f = 2 * sin(theta_h) * cos(theta_d); 
		for (int channel = 0; channel < 3; channel++) {
			double l2 = lambda[channel] * lambda[channel];
			// Multiplicative factor for specular reflection
			// A = portion of energy going into specular beam
			double A_specular = 1.0, A_diffraction = 0.0;
			double Q = 0;
			if (useDiffraction) {
				// GGX/ABC for diffraction part:
				double al = lambda[channel] / bdiff; 
				double u = cos(theta_d) / al;
				double scale = 1.0; 
				bool isDirac = false; 
				double u_prime = convolutionResult(u, b, p, c, scale, isDirac); 
				double f = 2 * sin(theta_h) * u_prime;
				double f2 = f * f;
				double sigma_rel2_over_sigma_s2 = 1.0; 
				if (isDirac) sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_i), c);
				else sigma_rel2_over_sigma_s2 = sigma_rel2(al * al, sin(theta_d), c);
				// A_spec = portion of energy going into specular beam
				double exp_term = 2. * M_PI * sigma_s * 2 * cos(theta_d);
				exp_term *= exp_term;
				A_specular = exp(- exp_term * sigma_rel2_over_sigma_s2/ l2);
				// A_diff = portion of energy going into diffraction beam
				if (isDirac) {
					exp_term = 2. * M_PI * sigma_s * (cos(theta_i) + cos(theta_o));
					exp_term *= exp_term;
					A_diffraction = exp(- exp_term * sigma_rel2_over_sigma_s2 / l2);
					Q = fresnelQ(ior[channel], 0.0, theta_i, theta_o, phi_o);
					scale = 1; // should be useless
				} else {
					// Multiplicative factor for specular reflection
					// A_specular = 1 - exp_term * sigma_rel2_over_sigma_s2 / l2;
					// if (A_specular < 0) A_specular = 0;
					A_diffraction = A_specular;
					Q = fresnelQ(ior[channel], 0.0, theta_d, theta_d, M_PI);
				}
				// GHS with ABC PSD
				diffraction = ((1. - A_diffraction) * Q * scale * (c - 1.)/ (2. * M_PI * al * al)) * pow(1 + f2, - 0.5 * (c+1));
				// re-normalization:
				diffraction /= sigma_rel2_over_sigma_s2;
			}
			double F = fresnel4(ior[channel], 0.0, cos(theta_d));
			// diffuse layer underneath, formula from Mitsuba plastic.cpp, with nonlinear = true:
			double F_dr_internal = Fdr(1.0/ior[channel]);
			// Single scattering BRDF 
			double cos_theta_o_inside = sqrt(1. - (sin(theta_o)/ior[channel]) *  (sin(theta_o)/ior[channel]));
			// double cos_theta_i_inside = sqrt(1. - (sin(theta_i)/ior[channel]) *  (sin(theta_i)/ior[channel]));
			double diffuse = albedo[channel] * (1. - fresnel4(ior[channel], 0.0, cos(theta_i))) * (1. - fresnel4(1./ior[channel], 0.0, cos_theta_o_inside)) / (M_PI * /*ior[channel] * ior[channel] */ (1.0 - albedo[channel] * F_dr_internal)); 
			// For future reference: factor 1/pi. Make sure is present in Mitsuba.
			std::cout << mesured[channel] << " " << estimate[channel] << " " << F * A_specular * cookTorrance + diffuse << " "  << F * A_specular * cookTorrance << " " << G2val * diffraction << " " << diffuse << " ";
		}
		std::cout << std::endl;
	}
	std::cout << std::endl;
	std::cout << std::endl;
}


void readData(const char* filepath)
{	
  // read brdf from file
  if (!MERLBRDF::read_brdf(filepath, brdfMesured)) 
  {
    fprintf(stderr, "Error reading %s\n", filepath);
    exit(1);
  }

  skipReading[0] = 0; 
  skipReading[1] = BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D/2;
  skipReading[2] = BRDF_SAMPLING_RES_THETA_H*BRDF_SAMPLING_RES_THETA_D*BRDF_SAMPLING_RES_PHI_D;
  scaleReading[0] = RED_SCALE;
  scaleReading[1] = GREEN_SCALE;
  scaleReading[2] = BLUE_SCALE; 
  // Theoretical RGB values 
  lambda[0] = 0.645; 
  lambda[1] = 0.526; 
  lambda[2] = 0.444; 
  // Retiga 1300 values from the internet
  // works, too, but you have to remove the scaling in MERLBRDFRead.h
  // lambda[0] = 0.610; 
  // lambda[1] = 0.540; 
  // lambda[2] = 0.470; 

  //  From Bagher et al. TOG, 2016: weight samples according to relation to median BRDF value.
  //  Weighted by importance (wv and wi). This is just to get the median value
  computeMedian(brdfMesured, median); 

  const int maxDataSize = 12; // Max data size. 8 for fit without k, 11 for fits with k.
  const int maxNumSteps = 800; 
  const int numTestsTogether = 10;
  double errorsTogether[numTestsTogether];
  double solutionsTogether[numTestsTogether][maxDataSize];
  double initialData[maxDataSize];
  double lowerBounds[maxDataSize];
  double upperBounds[maxDataSize];
  double typicalValue[maxDataSize];
  double measurement[3 * BRDF_SAMPLING_RES_THETA_H * BRDF_SAMPLING_RES_THETA_D];
  double *measurementTogether;
  double info[LM_INFO_SZ];
  double opts[5];
  bool showResults = true;
  // 1st test: all together, over entire range
  double* dataForFitting;
  double errorMin;
  int kmin;
  int numErrorMinFound;

  N_theta_i = 15; // Number of tests in theta_i;
  N = 90;
  d_theta_i = cutoff_theta_i / N_theta_i; 
  skip = 2 * N * N * N_theta_i; 
  measurementTogether = new double[3 * skip];

  dataForFitting = initialData;
  for (int k = 0; k < numTestsTogether;) {
    for(int i=0; i< 3 * skip; ++i)
      measurementTogether[i] = 0.0;
    for (int i=0; i < LM_INFO_SZ; i++) info[i] = 0.0;
    opts[0] = 1.0;
    // These are easier options
    opts[1] = 1e-15;
    opts[2] = 1e-15;
    opts[3] = 1e-15;
	opts[4] = -1e-9;
    int dataSize = 0;
	
	// Transparent but colored layer
    initialData[dataSize] = 1. + 0.2 * drand48(); 
    lowerBounds[dataSize] = 1.;
    upperBounds[dataSize] = 2.0;
    typicalValue[dataSize] = 1.;
    dataSize += 1;
    initialData[dataSize] = initialData[0]; 
    lowerBounds[dataSize] = 1.;
    upperBounds[dataSize] = 2.0;
    typicalValue[dataSize] = 1;
    dataSize += 1;
    initialData[dataSize] = initialData[0]; 
    lowerBounds[dataSize] = 1.;
    upperBounds[dataSize] = 2.0;
    typicalValue[dataSize] = 1.;
    dataSize += 1;

	// shape parameter values:
    if (useDiffraction) {
    	// characteristic length b, in micro-meters. 
    	// larger than wavelength (0.3), smaller than 10 mum (10)
    	initialData[dataSize] = 0.5 + 0.3 * drand48(); 
    	lowerBounds[dataSize] = 0.1; // ???
    	upperBounds[dataSize] = 200; // ??? 
    	typicalValue[dataSize] = 0.5;
    	dataSize += 1;
	}
    if (useDiffraction) {
    	initialData[dataSize] = 1.1 + drand48(); // c 
    	lowerBounds[dataSize] = 1.02; // 1 is mandatory
    	upperBounds[dataSize] = 30.02; // ??? Max convolution computed for 3.02. Beyond that, unknown
    	typicalValue[dataSize] = 1.3;
    	dataSize += 1;
	}
    if (useReflection) {
#define Shiny
#ifdef Shiny
		// b, average slope. Between 0 and 1 (relative to bMax)
		initialData[dataSize] = 0.01 + 0.02 * drand48();      // metals
    	lowerBounds[dataSize] = 1e-6; 
    	upperBounds[dataSize] = 1.0; 
    	typicalValue[dataSize] = 0.001; // metals
    	dataSize += 1;
		initialData[dataSize] = 0.2 + 0.1 * drand48(); // p 
		lowerBounds[dataSize] = 0.05;  
		upperBounds[dataSize] = 2.0;
		typicalValue[dataSize] = 0.2;
		dataSize += 1;
#else
		initialData[dataSize] = 0.2 + 0.2 * drand48();   
    	lowerBounds[dataSize] = 0.1; 
    	upperBounds[dataSize] = 1.0; 
    	typicalValue[dataSize] = 0.5;
    	dataSize += 1;
		initialData[dataSize] = 0.8 + 0.9 * drand48(); 
		lowerBounds[dataSize] = 0.5;  
		upperBounds[dataSize] = 5.0;
		typicalValue[dataSize] = 0.8;
		dataSize += 1;
#endif
	}
#undef Shiny
#ifdef Shiny
  	double sigmaList[] = {0.03, 0.03, 0.01, 0.01, 0.001};
#else
	// double sigmaList[] = {1.0, 1.0, 0.3, 0.3, 0.5};
  	double sigmaList[] = {0.03, 0.03, 0.01, 0.01, 0.001};
#endif
	if (useDiffraction) {
		int step = (k / 2) % 5; // 0, 0, 1, 1, 2, 2, ..., 4
		double defaultValue = sigmaList[step];
    	initialData[dataSize] = defaultValue + defaultValue * drand48(); // sigma_s
    	lowerBounds[dataSize] = 0.;
    	upperBounds[dataSize] = 2;
    	typicalValue[dataSize] = defaultValue;
    	dataSize += 1;
    }

	int reference = dataSize;
	initialData[dataSize] = 0.01 * drand48();        // albedo
	lowerBounds[dataSize] = 0.0;
	upperBounds[dataSize] = 1.0;
	typicalValue[dataSize] = 0.01;
	dataSize += 1;
	initialData[dataSize] = initialData[reference]; // albedo
	lowerBounds[dataSize] = 0.0;
	upperBounds[dataSize] = upperBounds[reference];
	typicalValue[dataSize] = typicalValue[reference];
	dataSize += 1;
	initialData[dataSize] = initialData[reference]; // albedo
	lowerBounds[dataSize] = 0.0;
	upperBounds[dataSize] = upperBounds[reference];
	typicalValue[dataSize] = typicalValue[reference];
	dataSize += 1;

	for (int h = dataSize; h < maxDataSize; h++) initialData[h] = 0.0;

    printf("# Fitting all parameters together. k= %d\n# Initial data = ", k);
    for(int i=0; i < dataSize; ++i) {
      printf("%.7g ", initialData[i]);
    }
    printf("\n");
    int ret;
#define TWO_PASS
#ifdef TWO_PASS
    // LevMar, Box constraints, Analytic differentials
	ret = dlevmar_bc_dif(measure1dimension, initialData, measurementTogether, 
			dataSize, 3 * BRDF_SAMPLING_RES_THETA_H, 
			lowerBounds, upperBounds, typicalValue, 
			4000, opts, info, NULL, NULL, NULL);
	printf("# Levenberg-Marquardt 1D returned %d in %g iter, reason %g, error = %g, k= %d\n", ret, info[5], info[6], info[1], k);
	fflush(stdout);

	// Only increment k if 1st step converged and E < 1e-3
	// Otherwise, retry. More 1D tests, less useless 2D tests
	if ((info[6] <= 3) && (info[1] < 0.1)) {
		for (int i = 0; i < maxDataSize; i++) solutionsTogether[k][i] =  initialData[i];
		if (showResults) displaySolution(initialData); 
		for (int h = 0; h < dataSize; h++) {
			if (initialData[h] > 0) typicalValue[h] = initialData[h];
		}
#endif
		ret = dlevmar_bc_dif(measureSpherical, initialData, measurementTogether, 
				dataSize, 3 * skip, 
				lowerBounds, upperBounds, typicalValue, 
				20, opts, info, NULL, NULL, NULL);
		printf("# Levenberg-Marquardt 2D returned %d in %g iter, reason %g, error = %g, k= %d\n", ret, info[5], info[6], info[1], k);
		fflush(stdout);

		errorsTogether[k] = info[1]; // Error by fitting
		for (int i = 0; i < maxDataSize; i++) solutionsTogether[k][i] =  initialData[i];
		if (showResults) displaySolution(solutionsTogether[k]); 
		k += 1;
#ifdef TWO_PASS
	} else errorsTogether[k] = HUGE_VAL;
#endif
  }
  errorMin = HUGE_VAL;
  kmin = -1;
  numErrorMinFound = 0;
  for (int k = 0; k < numTestsTogether; k++) {
    if (errorsTogether[k] < errorMin) {
      errorMin = errorsTogether[k];
      kmin = k;
    }
  }
  if (kmin >= 0) {
    for (int k = 0; k < numTestsTogether; k++) {
      if (errorsTogether[k] < errorMin * 1.1) numErrorMinFound++;
    }
    std::cout << "# Best solution for Together = " << kmin << " error = " << errorMin << ", found " << numErrorMinFound << " times out of " << numTestsTogether << " (" << (float)numErrorMinFound*100./numTestsTogether << " %)"<< std::endl;
    // if (!showResults) displaySolution(solutionsTogether[kmin]); 
  }

  free(brdfMesured);
  delete [] measurementTogether;
}


int main(int argc, char* argv[])
{
  if(argc < 2) {
    std::cerr << "Usage: ./a.out brdf_name" << std::endl;
    exit(1);
  }
  const char *input_filename = argv[1];

  readData(input_filename);
  return 0;
}

