/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */
//
// Code for computing the convolution between the normal probability distribution D and the diffraction lobe S. 
// Relies on the Levenberg-Marquadt implementation levmar: http://users.ics.forth.gr/~lourakis/levmar/ and on the GNU Scientific Library
// You'll need to specify the path to levmar.h and the GSL and levmar.a libraries

// How to compile: 
// On Linux: g++ -O3 -mavx -I/usr/local/include -L/usr/local/lib computeConvolution.cpp -lm -lgsl  -I../unifiedModel/ ./levmar-2.6/liblevmar.a -llapack -lcblas -lblas -I./levmar-2.6/
// On a Mac:
//  clang++ -Wall -O3 -mavx  -I/usr/local/include -L/usr/local/lib computeConvolution.cpp -lm -lgsl  -I../unifiedModel ./levmar-2.6/liblevmar.a -llapack -lcblas -I/opt/local/include/ --std=c++11 -L/opt/local/lib/ 

// How to execute: ./a.out > ctd_convolution2.h

/**
 *  Nicolas Holzschuch @ inria.fr
 *  Romain Pacanowski @ institutoptique.fr
 **/

#include <iostream>
#include <istream>
#include <fstream>
#include <cstdlib>
//#include <math.h>
#include "MERLBRDFRead.h"
#include "common_functions.inl"
#include <complex.h> 
#include <gsl/gsl_sf_legendre.h>
// This is where you need to give the path to levmar.h
#include "levmar-2.6/levmar.h"

#include <cmath>
#include <cstring>
#include <vector>


// g++ -I/usr/local/include -L/usr/local/lib computeConvolution4.cpp -lm -lgsl  ../../../levmar-2.6/liblevmar.a -llapack -lcblas 

double b_l_cos_td;
double c;
double beta;
double p; 

// Fitting with D(0) * Sz(u') == 2 parameters, u' and a1
// Fits roughly well. 
// Dirac ==> u' = u, a1 = pi/numZHcoeffs0 (why ???) == dx ?
// Not Dirac ==> u' increases, then stops. a1 = other law.
// Simple enough to be useable. Outliers for p = 1 and 5 (to be removed)

// delta_p = 0.05
double betaLimits[98] = {0.058, 0.059, 0.059, 0.06, 0.06, 0.06, 0.061, 0.061, 0.062, 0.062, 0.063, 0.063, 0.064,
	0.064, 0.065, 0.065, 0.066, 0.066, 0.067, 0.068, 0.068, 0.069, 0.07, 0.07, 0.071, 0.072, 
	0.073, 0.074, 0.074, 0.075, 0.076, 0.077, 0.078, 0.079, 0.081, 0.082, 0.083, 0.084, 0.084, 
	0.085, 0.085, 0.086, 0.086, 0.087, 0.087, 0.088, 0.089, 0.089, 0.09, 0.091, 0.092, 0.092, 
	0.093, 0.094, 0.095, 0.096, 0.098, 0.099, 0.1, 0.102, 0.104, 0.106, 0.108, 0.11, 0.113, 0.116,
	0.119, 0.123, 0.127, 0.132, 0.138, 0.145, 0.153, 0.163, 0.175, 0.191, 0.211, 0.236, 0.268, 0.308, 
	0.348, 0.388, 0.416, 0.441, 0.461, 0.472, 0.47, 0.469, 0.448, 0.418, 0.389, 0.327, 0.261, 0.195, 
	0.121, 0.055, 0.015, 0.001};

// Summary: Coeffs SH pr ABC = a * exp(-(x/b)^p) . Coeffs SH pr ExtGauss = oscillations. Fréquence ? Décroissance ? 
// What happens for ext Gauss when p = 0.1 ? violent
// ABC = identifier patterns. b_l_cos_td > 50 = quasi-Dirac --> ? 
//
// Modèle, modèle, modèle. 
// No model, just a table.
// u + (c, beta, p) --> new u
// Fitting with LevMar behaving badly
// 1) find best blc
// 2) blc fixed, find best c
// 3) iterate (scale comes from computations)
//
// 3 parameters: scale1, u' (from u), scale2
// Variable transfer?
// 


void errorFunctionValue(double *params, double *measurement, int dataSize, int numSamples, void *adata)
{
	double *value_ref = (double*) adata;
	double sum = 0;
	double scale1 = (params[0]); 
	double uprime = params[1]; 
	// Multiplicative factor for specular reflection

#pragma omp parallel for
	for (int i = 0; i < numSamples; i++) {
		double sin_x = 1. - ((double) i ) / numSamples;
		double cos_x = sqrt(1. - sin_x * sin_x);
		double f2 = 4 * (1 - cos_x * cos_x) * uprime * uprime;
		double funcVal = scale1 * ((p  * gsl_sf_gammainv(1./p)) /(beta * beta * M_PI)) * pow(1 + f2, - 0.5 * (c+1));
		measurement[i] = fabs(value_ref[i] - funcVal);	
	}		
}

double fitModelValue(double* coeffs, int numSamples, double *params) {
	double info[LM_INFO_SZ];
	double opts[5];

	double measure[numSamples]; 
	for(int i=0; i< numSamples; ++i) measure[i] = 0.0;
	for (int i=0; i < LM_INFO_SZ; i++) info[i] = 0.0;
	opts[0] = 1.0;
	// These are easier options
	opts[1] = 1e-15;
	opts[2] = 1e-15;
	opts[3] = 1e-15;
	opts[4] = -1e-9;

	int ret = dlevmar_dif(errorFunctionValue, params, measure, 
			2, numSamples, 100, opts, info, NULL, NULL, coeffs);
	return info[1]; 
}




// This number of samples is the same as the one defined by Nicolas
int const num_of_P = 100;
int const num_of_Beta = 10;
int const num_of_b_l_cos_td = 43;
int const num_of_c  = 10;

std::string output_result_FILENAME = "convolution_results.txt";




struct CTDConvolutionResults
{
	double scale;
	double uprime;
};


bool saveResultsOnDisk( std::string const & filename,
		// std::vector<double> const & p_values,
		// std::vector<double> const & beta_values,
		// std::vector<double> const & b_l_cos_td_values,
		// std::vector<double>const  & c_values,
		CTDConvolutionResults res_fit[num_of_P][num_of_Beta][num_of_b_l_cos_td][num_of_c] )
{
	using namespace std;

	ofstream file_writer( filename );

	if( !file_writer )
	{

		return false;
	}

	file_writer << "#ALTA_BEGIN_HEADER"  << std::endl;
	file_writer << "#DIM 4 2"  << std::endl;
	file_writer << "#PARAM_IN UNKNOWN "  << std::endl;
	file_writer << "#PARAM_OUT UNKNOWN "  << std::endl;
	file_writer << "#VS 0"  << std::endl;
	file_writer << "#ALTA_END_HEADER"  << std::endl;

	for( unsigned int i=0; i < num_of_P; i++)
	{


		for( unsigned int j=0; j < num_of_Beta; j++)
		{
			for( unsigned int k=0; k < num_of_b_l_cos_td; k++)
			{
				for( unsigned int l=0; l < num_of_c; l++)
				{
					file_writer << (res_fit[i][j][k][l]).scale <<  " " << res_fit[i][j][k][l].uprime ;
				}
			}
		}

		file_writer << endl;
	}

	file_writer.close();
	return true;
}



// SOME CONSTANTS TO SPEED-UP 
double const TWO_PI = 2.0 * M_PI;
double const FOUR_PI = 4.0 * M_PI;


int main(int argc, char** argv) 
{
	using namespace mrf::util;

	int const N = 600000; // in order to get inverse SH transform accurate, need enormous number of samples. 

	// Not so much if only interested in direct SH transform
	int const numZHcoeffs = 100; 

	double coeffs_ABC[numZHcoeffs];
	double coeffs_EG[numZHcoeffs];
	double coeffs_product[numZHcoeffs];

	std::vector<double> cte_coeffs_ZH;
	std::vector<double> two_pi_cte_coeffs_ZH;
	std::vector<double> inv_cte_coeffs_ZH;


	// PRE-COMPUTE SH NORMALIZATION CONSTANTS
	for( unsigned int l=0; l < numZHcoeffs; l++)
	{

		double const normalization_cte =  sqrt((2 * l +1 )/(FOUR_PI));
		cte_coeffs_ZH.push_back( normalization_cte );
		two_pi_cte_coeffs_ZH.push_back( TWO_PI * normalization_cte );

		inv_cte_coeffs_ZH.push_back( sqrt(FOUR_PI/(2 * l +1 ) ) );
	}

	double const dx = 1./N;


	// Pre-compute gamma_inv(1/p)
	std::vector<double> gamma_inv_1_p ;
	std::vector<double> p_values;

	for( p = 5.0; p > 0.0; p-= 0.2)
	{
		gamma_inv_1_p.push_back( p *  gsl_sf_gammainv( 1. / p ) );

		p_values.push_back( p );
		std::cout << " p = " <<  p << std::endl;
	} 

	std::vector< std::vector<double> > beta_values;
	beta_values.resize( p_values.size() );
	for(unsigned int index_p=0; index_p < p_values.size(); index_p++)
	{

		double const maxBeta = betaLimits[index_p*4];
		double const deltaBeta = 0.1 * maxBeta;
		for( double current_Beta=0.05 * maxBeta; current_Beta < maxBeta; current_Beta += deltaBeta )
		{
			beta_values[index_p].push_back( current_Beta);
		}
	}

	std::cout << "p_values.size() = " << p_values.size() << std::endl;

	//c = 3.02 ; c > 1; c -= 0.2) 
	std::vector<double> c_values;
	for( c=3.02 ; c > 1. ; c -= 0.2)
	{
		c_values.push_back( c );
	}
	std::cout << " c_values.size() = " << c_values.size() << std::endl;
	for(unsigned int i=0; i < c_values.size(); i++)
	{
		std::cout << " c_values[" << i <<"] = " << c_values[i] << std::endl;
	}


	// PRECOMPUTING b_l_cos_td values 
	std::vector<double> b_l_cos_td_values;

	double delta_blc = 0.05; 
	for (b_l_cos_td = 0; b_l_cos_td <= 100.0; b_l_cos_td +=  delta_blc) 
	{
		if (b_l_cos_td <= 1) delta_blc = 0.05;
		else if (b_l_cos_td < 10) delta_blc = 1;
		else if (b_l_cos_td < 50) delta_blc = 5.0;
		else delta_blc = 10.0;

		b_l_cos_td_values.push_back( b_l_cos_td );
		// std::cout << " b_l_cos_td = " << b_l_cos_td << std::endl;
	}
	std::cout << " b_l_cos_td_values.size() = " << b_l_cos_td_values.size() << std::endl;




	// PRE-COMPUTE legendre evaluation for ALL SH band and cosine evaluation
	std::vector< std::vector<double> > dx_legendre_cos_N_l;
	double* cos_x_at_N       = new double[N];
	double* sqr_cos_x_at_N   = new double[N];
	double* four_one_minus_cos_N = new double[N];
	double* sqr_tan_at_N     = new double[N];
	double* cos_x_pow_4_at_N = new double[N];


	dx_legendre_cos_N_l.resize( N );

#pragma omp parallel for
	for( unsigned int i=0; i < N; i++)
	{
		cos_x_at_N[i] = 1. - ((double) i ) / N;
		sqr_cos_x_at_N[i] = cos_x_at_N[i] * cos_x_at_N[i];
		four_one_minus_cos_N[i] = 4 * (1 - sqr_cos_x_at_N[i]);
		cos_x_pow_4_at_N[i] =  sqr_cos_x_at_N[i] *  sqr_cos_x_at_N[i];

		sqr_tan_at_N[i] = (1. - sqr_cos_x_at_N[i]) /  sqr_cos_x_at_N[i];

		for( unsigned int l=0; l< numZHcoeffs;  l++)
		{
			dx_legendre_cos_N_l[i].push_back(  dx * gsl_sf_legendre_Pl(l, cos_x_at_N[i] ) );
		}
	}

	CTDConvolutionResults  res_fit[p_values.size()][num_of_Beta][b_l_cos_td_values.size()][c_values.size()];


	// Precision of the convolution
	const int numSamples = 1000; 
	double cos_x_at_NumSamples[numSamples];
	double sin_x_at_NumSamples[numSamples];

	double sqr_cos_x_at_NumSamples[numSamples];

	std::vector< std::vector<double> > legendre_cos_NumSamples_l;

	legendre_cos_NumSamples_l.resize(numSamples);

	for (unsigned int i = 0; i < numSamples; ++i)
	{
		sin_x_at_NumSamples[i] = 1. - ((double) i ) / numSamples;
		cos_x_at_NumSamples[i] = std::sqrt( 1. -  sin_x_at_NumSamples[i]*sin_x_at_NumSamples[i]);

		sqr_cos_x_at_NumSamples[i] = cos_x_at_NumSamples[i] * cos_x_at_NumSamples[i];

		for( unsigned int l=0; l < numZHcoeffs; l++)
		{
			legendre_cos_NumSamples_l[i].push_back( gsl_sf_legendre_Pl(l, cos_x_at_NumSamples[i]) );
		}

	}




	// NumSamples : p = 100, beta = 10, b_l_cos_td = 43, c = 10 : 
	// Loop on Extended Gaussian parameters
	// for (p = 0.2; p <= 5.; p+= 0.2) 
	// i_p = (5 - p) * 20
	int i_p = 0; 

	unsigned int number_of_values_computed = 0;

	// for (p = 5.0, i_p = 0; p >= 0.2; p-= 0.2, i_p += 4) 
	//#pragma omp parallel for default(none)
	for( unsigned int index_p = 0; index_p < p_values.size(); index_p++)
	{

		std::cout << " p = " << p << " i_p = " << i_p << std::endl;

		p = p_values[index_p];


		//double const maxBeta = betaLimits[i_p];
		double const maxBeta = betaLimits[index_p*4];

		double const deltaBeta = 0.1 * maxBeta;

		//double const p_gamma_inv_one_p =  p  * gsl_sf_gammainv(1./p) ;
		double const p_gamma_inv_one_p =  p_values[index_p]  * gsl_sf_gammainv(1./p_values[index_p]) ;

		//#pragma omp parallel for 
		for (beta = 0.05 * maxBeta;  beta < maxBeta; beta += deltaBeta)
		{
			//for (int l = 0; l < numZHcoeffs; l++) coeffs_EG[l] = 0.0;
			//std::fill_n( &coeffs_EG[0], numZHcoeffs, 0.0);
			memset( &coeffs_EG[0], 0, sizeof( coeffs_EG  ) );

			double const sqr_beta = beta * beta;
			double const PI_sqr_beta = sqr_beta*M_PI;

			for (int i = 0; i < N; i++) 
			{
				//double cos_x = 1. - ((double) i ) / N;
				//double c2 = cos_x * cos_x;
				double const c2 = sqr_cos_x_at_N[i];
				//double t2; 
				double funcVal; 

				if (c2 != 0) 
				{
					//double const t2 = (1 - c2)/c2;
					double const t2 = sqr_tan_at_N[i];
					//funcVal = p  * gsl_sf_gammainv(1./p) * exp(-pow(t2/(beta * beta), p)) /(beta * beta * M_PI * c2 * c2);
					funcVal =  p_gamma_inv_one_p * exp(-pow(t2/(sqr_beta), p)) /(PI_sqr_beta* cos_x_pow_4_at_N[i]);
				} else 
				{
					assert(0);
					funcVal = 0;
				}

				for (int l = 0; l < numZHcoeffs; l++) {
					//coeffs_EG[l] += TWO_PI * funcVal * sqrt((2 * l +1 )/(FOUR_PI)) * gsl_sf_legendre_Pl(l, cos_x) * dx;
					//coeffs_EG[l] += funcVal * two_pi_cte_coeffs_ZH[l] * gsl_sf_legendre_Pl(l, cos_x) * dx;
					coeffs_EG[l] += funcVal * two_pi_cte_coeffs_ZH[l] * dx_legendre_cos_N_l[i][l];
				}
			}

			// We have the SH coefficients for the Extended Gaussian

			// Loop over ABC parameters : from 3 to 1.02
			//for (c = 3.02 ; c > 1; c -= 0.2) 
			//#pragma omp parallel for default(none)
			for( unsigned int c_index=0; c_index < c_values.size(); c_index++)	
			{
				c = c_values[c_index];

				// Beyond that, close to a Dirac. Hard to fit, but continuous with previous terms
				// for (b_l_cos_td = 0.01; b_l_cos_td < 50.0; b_l_cos_td += 0.1) 
				double delta_blc = 0.05; 

				// for (b_l_cos_td = 50; b_l_cos_td > 0.0; b_l_cos_td -=  delta_blc) 
				// for (b_l_cos_td = 0; b_l_cos_td <= 100.0; b_l_cos_td +=  delta_blc) 
				// 	if (b_l_cos_td <= 1) delta_blc = 0.05;
				// 	else if (b_l_cos_td < 10) delta_blc = 1;
				// 	else if (b_l_cos_td < 50) delta_blc = 5.0;
				// 	else delta_blc = 10.0;

				for( unsigned int b_l_cos_td_index=0; b_l_cos_td_index < b_l_cos_td_values.size(); b_l_cos_td_index++)
				{
					b_l_cos_td = b_l_cos_td_values[ b_l_cos_td_index ];

					// for (b_l_cos_td = 1.2; b_l_cos_td > 0.0; b_l_cos_td -= 0.02) 
					//for (int l = 0; l < numZHcoeffs; l++) coeffs_ABC[l] = 0.0;
					//std::fill_n( &coeffs_ABC[0], numZHcoeffs, 0.0);beta * be
					memset( &coeffs_ABC[0], 0.0, sizeof( coeffs_ABC) );


					double const sqr_b_l_cos_td = b_l_cos_td * b_l_cos_td;

					//#pragma omp parallel for
					for (int i = 0; i < N; i++) 
					{
						//double cos_x = 1. - ((double) i ) / N;
						//double const f2 = 4 * (1 - cos_x * cos_x) * b_l_cos_td * b_l_cos_td;

						double const f2 = four_one_minus_cos_N[i] * sqr_b_l_cos_td;

						// Full function evaluation
						// double funcVal = b_l_cos_td * b_l_cos_td * cos_x * cos_x * ((c - 1.)/ (2. * M_PI )) * pow(1 + f2, - 0.5 * (c+1));
						// Pull out everything independent from cos_x: 

						//double const funcVal = cos_x * cos_x * pow(1 + f2, - 0.5 * (c+1));

						double const funcVal = sqr_cos_x_at_N[i] * pow(1 + f2, - 0.5 * (c+1) );

						for (int l = 0; l < numZHcoeffs; l++) 
						{
							//coeffs_ABC[l] += TWO_PI * funcVal * sqrt((2 * l +1 )/(FOUR_PI)) * gsl_sf_legendre_Pl(l, cos_x) * dx;
							coeffs_ABC[l] += two_pi_cte_coeffs_ZH[l] * funcVal  * dx_legendre_cos_N_l[i][l] ;
						}
					}

					// Now, we have the SH coefficients of the ABC function. Compute the convolution:
					for (int l = 0; l < numZHcoeffs; l++) {
						coeffs_product[l] = coeffs_ABC[l] * coeffs_EG[l] * inv_cte_coeffs_ZH[l];
					}


					// Now we have the convolution. 
					double convolved[numSamples]; 
					//for (int i = 0; i < numSamples; i++) convolved[i] = 0.; 
					memset( &convolved[0], 0.0, sizeof( convolved) );

					for (int i = 0; i < numSamples; i++) {
						//double sin_x = 1. - ((double) i ) / numSamples;
						//double cos_x = sqrt(1. - sin_x * sin_x);
						//double funcVal = 0; 

						for (int l = 0; l < numZHcoeffs; l++) {
							//convolved[i] += coeffs_product[l] * sqrt( (2 * l +1 )/(FOUR_PI)) * gsl_sf_legendre_Pl(l, cos_x);
							//convolved[i] += coeffs_product[l] * cte_coeffs_ZH[l] * gsl_sf_legendre_Pl(l, cos_x);
							convolved[i] += coeffs_product[l] * cte_coeffs_ZH[l] * legendre_cos_NumSamples_l[i][l];

						}
					}


					// FINAL PART
					double params[2]; 
					params[0] = 0.3;
					params[1] = b_l_cos_td;
					double error = fitModelValue(convolved, numSamples, params);
					std::cout << beta << " " << p << " " << b_l_cos_td << " " << c << " " 
						<< (params[0]) << " " << fabs(params[1]) << " " << error << std::endl;

					number_of_values_computed++; 
				} 
			}// end of loop for-c
		}//end of for-loop on beta
	}// end of for-loop on p

	delete[] cos_x_at_N;
	delete[] sqr_cos_x_at_N;
	delete[] cos_x_pow_4_at_N;
	delete[] sqr_tan_at_N;
}

