/**
 * Copyright (C) 2017 INRIA  http://www.inria.fr
 * Copyright (C) 2017 CNRS   http://www.cnrs.fr
 *
 * Authors:
 *  Nicolas Holzschuch:  nicolas.holzschuch@inria.fr
 *  Romain  Pacanowski:  romain.pacanowski@institutoptique.fr 
 *
 * Permission to use, copy and modify this software and its documentation without
 * fee for educational, research and non-profit purposes, is hereby granted, provided
 * that the above copyright notice appear in all copies.
 * 
 * To request permission to incorporate this software into commercial products contact
 * the authors above. 
 */

#include <iostream>
#include <fstream>
#include <math.h>
#include "../unifiedModel/common_functions.inl"
#include <complex.h> 
#include <fftw3.h>

// Code to precompute the effect of the convolution by camera sensor on distribution
// convolving camera sensor with distribution D, resulting in low-pass filter
// We approximate the effect of that filter as reducing intensity (it also smoothes the signal, 
// but we try to limit the number of parameters). 
// Intensity is reduced using ((1 - x*2/pi)(1 + x*2/pi))**power 

// Compile with: g++ precomputeConvolution.cpp -lm -lgsl -lfftw3 -o convolution
// Execute with: ./a.out > precomputedLowPass.h

double power[100][1000]; // power values, indexed by kp = 5./p and ks = exp(-u^(1/20))
// data allocation outside the function for efficiency
fftw_complex *in, *out;
fftw_plan p_in, p_out;


double error(double params, double* inputData, int numSamples) {
	double returnValue = 0;
	for (int i = 0; i < numSamples; i++) {
		double theta_o = (0.5 * M_PI * i)/numSamples;
		double poly = (1 - theta_o * 2./M_PI) * (1 + theta_o * 2./M_PI);
		poly = pow(poly, params); 
		assert(!isnan(poly));
		if (theta_o < 1.2) 
			returnValue += (inputData[i] - poly) * (inputData[i] - poly);
	}
	return returnValue;
}

inline double derivative(double params, double* inputData, int numSamples) {
	double h = 1e-6;
	return (error(params+h, inputData, numSamples) -  error(params-h, inputData, numSamples))/h;
}

double bestParameter(double* inputData, int numSamples) {
	double alpha = 0.5;
	double stepSize = 0.1;

	double deriv = derivative(alpha, inputData, numSamples);
	double direction = deriv > 0 ? -1.0 : 1.0;
	bool keepGoing = true;
	do {
		do {
			alpha += stepSize * direction;
			double newDeriv = derivative(alpha, inputData, numSamples);
			keepGoing = (newDeriv * direction) < 0; 
			// std::cerr << alpha << " " << error(alpha, inputData, numSamples) << " " << stepSize << std::endl;
		} while (keepGoing && (alpha > 0)); 
		if (!keepGoing) {
			direction *= -1; 
			stepSize /= 2;
		} else {
			alpha = 0.5 * stepSize; 
			stepSize /= 4; 
			direction = derivative(alpha, inputData, numSamples) > 0 ? -1.0 : 1.0;
		}
	} while (stepSize > 1e-5); 

	return alpha; 
}


double estimateAlpha(double p, double beta) {
	double sigma_r = (double) N / 256.; // 512 pixels over the sphere
	int numSamples = 157;
	double inputData[numSamples]; 
	double f00;

	for (int k = 0; k < numSamples; k++) {
		double theta_o = (0.5 * M_PI * k)/numSamples;
		// load the values. These don't depend on theta_o
		// but we recompute anyway since it was erased by the double transform
		for (int i = 0; i <= N/2; i++) {
			double theta_h = 0.5 * M_PI * (double) i / N;
			in[i][0] = D(theta_h, beta, p); // we only filter D, as it has most HF.
			if (i > 0) in[N - i][0] = D(theta_h, beta, p); 
			in[i][1] = 0.0;
			if (i > 0) in[N - i][1] = 0.0;
		}
		
		double one_sigma_p = cos(theta_o) / sigma_r;

		// FFT of D(theta_h) :
		fftw_execute(p_in); 
		// Now multiply by the gaussian
		for (int i = 0; i <= N/2; i++) {
			double omega = (double) i / N;
			double exp_term = omega / one_sigma_p ;
			double gaussian = exp( - exp_term * exp_term);
			out[i][0] *= gaussian;
			if (i > 0) out[N-i][0] *= gaussian;
			out[i][1] *= 0;
			if (i > 0) out[N-i][1] *= 0;
		}
		// Inverse Fourier Transform
		fftw_execute(p_out); 
		// We have the result of the convolution inside out
		// Now we approximate it.

		inputData[k] = in[0][0]/N;
		if (k == 0) f00 = inputData[0];
		inputData[k] /= f00;
	}
	// Now, we fit the data:
	double alpha = bestParameter(inputData, numSamples); 

	return alpha; 
}

int main() {
	N = 8192;
	in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
	out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
	p_in = fftw_plan_dft_1d(N, in, out, FFTW_FORWARD, FFTW_ESTIMATE);
	p_out = fftw_plan_dft_1d(N, out, in, FFTW_FORWARD, FFTW_ESTIMATE);

	std::cout << "precomputedPower[100][1000] = {\n{";
	// Now compute  power:
	for (int kp = 100; kp > 0; kp--) {
	// for (int kp = 25; kp > 0; kp--) {
		double p = 5./kp;
		double exponent = 2; 
		for (int kks0 = 0; kks0 < 1000; kks0++) {
			double ks0 = 1000 - kks0;
			double beta = 1./(ks0 * ks0);
			double alpha = estimateAlpha(p, beta);
			// std::cout << p << " " << beta << " " << alpha << std::endl;
			std::cout << alpha << ", " ;
		}
		std::cout << "},\n {";
		// std::cout << std::endl;
	}
	std::cout << "};" << std::endl;
	fftw_destroy_plan(p_in);
	fftw_destroy_plan(p_out);
	fftw_free(in); fftw_free(out);
}
